/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl;

import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.wordsource.WordSourceId;

import com.topcoder.util.compression.Base64Codec;
import com.topcoder.util.compression.Base64Decoder;
import com.topcoder.util.compression.Base64Encoder;
import com.topcoder.util.config.ConfigManager;
import com.topcoder.util.config.UnknownNamespaceException;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;

import java.lang.reflect.InvocationTargetException;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import java.util.zip.DataFormatException;


/**
 * <p>
 * Helper class for the Document Index Persistence Component.
 * </p>
 *
 * <p>
 * Note, since this is a multi-package component, this helper class is defined as public instead of package-private
 * to avoid defining such class in every package. This is allowed by the TopCoder rule.
 * All the static methods are shared between multiple packages are defined as public, otherwise will be defined as
 * package-private.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 *
 */
public final class DocumentIndexPersistenceUtil {

    /**
     * <p>
     * Constant used in byte2hexString().
     * </p>
     */
    private static final int CONSTANT = 0xFF;

    /**
     * <p>
     * Constant used in byte2hexString().
     * </p>
     */
    private static final int HEX = 16;

    /**
     * <p>
     * This private constructor prevents to create a new instance.
     * </p>
     */
    private DocumentIndexPersistenceUtil() {
    }

    /**
     * <p>
     * Checks whether the given Object is null.
     * </p>
     *
     * @param arg the argument to check
     * @param name the name of the argument to check
     *
     * @throws IllegalArgumentException if the given Object is null
     * @return the original object
     */
    public static Object checkNull(Object arg, String name) {
        if (arg == null) {
            throw new IllegalArgumentException(name + " should not be null.");
        }

        return arg;
    }

    /**
     * <p>
     * Checks whether the given String is null or empty.
     * Here empty means the length of the given string is zero after trimmed.
     * </p>
     *
     * @param arg the String to check
     * @param name the name of the String argument to check
     *
     * @throws IllegalArgumentException if the given string is null or empty
     * @return the original string
     */
    public static String checkString(String arg, String name) {
        checkNull(arg, name);

        if (arg.trim().length() == 0) {
            throw new IllegalArgumentException(name + " should not be empty.");
        }

        return arg;
    }

    /**
     * <p>
     * Checks whether the given arg represents an integer and its value satisfies the given operator compared to
     * the given value.
     * </p>
     *
     * <p>
     * Note, in current version, the given operator can be only "GE" or "GT".
     * "GE" means "greater than or equal to" while "GE" means "greater than".
     * </p>
     *
     * @param arg the String which represents an integer to check
     * @param name the name of the String argument to check
     * @param operator "GE"(&gt;=) or "GT"(&gt;)
     * @param value the given int value to be compared with
     * @return the int value that the given arg represents
     *
     * @throws PersistenceConfigurationException the given arg fails to satisfies the given operator compared to
     * the given value or when the given arg cannot be parsed to an integer
     */
    public static int checkInt(String arg, String name, String operator,
        int value) throws PersistenceConfigurationException {
        try {
            return checkInt(Integer.parseInt(arg), name, operator, value);
        } catch (NumberFormatException nfe) {
            throw new PersistenceConfigurationException(
                "The configed value for " + name + " is not an integer.",
                nfe);
        } catch (IllegalArgumentException e) {
            throw new PersistenceConfigurationException("The configed value for " + name + "is not valid.");
        }
    }

    /**
     * <p>
     * Checks whether the given int arg satisfies the given operator compared to the given int value.
     * </p>
     *
     * <p>
     * Note, in current version, the given operator can be only "GE" or "GT".
     * "GE" means "greater than or equal to" while "GE" means "greater than".
     * </p>
     *
     * @param arg the integer to check
     * @param name the name of the integer argument to check
     * @param operator "GE"(&gt;=) or "GT"(&gt;)
     * @param value the given int value to be compared with
     * @return the original int value
     *
     * @throws IllegalArgumentException the given arg fails to satisfies the given operator compared to
     * the given value
     */
    public static int checkInt(int arg, String name, String operator, int value) {
        // GT operator means greater than
        if ("GT".equalsIgnoreCase(operator)) {
            if (!(arg > value)) {
                throw new IllegalArgumentException(name
                        + " should greater than " + value);
            }
        // GE operator means greater than or equal to
        } else if ("GE".equalsIgnoreCase(operator)) {
            if (!(arg >= value)) {
                throw new IllegalArgumentException(name
                        + " should greater than or equal to " + value);
            }
        } else {
            // This will not happened.
            throw new IllegalArgumentException("Not supported operator.");
        }

        return arg;
    }

    /**
     * <p>
     * Return the value of the property.
     * </p>
     *
     * @param namespace the namespace to get
     * @param propertyName the name of property
     * @param required whether this property is required
     *
     * @return the value of the property
     *
     * @throws PersistenceConfigurationException if fail to load the config values
     */
    public static String getString(String namespace, String propertyName,
        boolean required) throws PersistenceConfigurationException {
        String property = null;

        try {
            property = ConfigManager.getInstance().getString(namespace,
                    propertyName);
        } catch (UnknownNamespaceException e1) {
            throw new PersistenceConfigurationException("Unknown namespace: "
                    + namespace, e1);
        }

        // This property is missed or the value empty
        if ((property == null) || (property.trim().length() == 0)) {
            if (required) {
                throw new PersistenceConfigurationException("Property "
                        + propertyName + "is missing.");
            }
        }

        return property;
    }

    /**
     * <p>
     * Uses the constructor specified by the parameters argument to create a new instance of the class of the given
     * className. The number of elements in the types array and parameters array must be equal.
     * </p>
     *
     * @param classType the object class type
     * @param className the name of the user defined class.
     * @param types the parameter types to the user defined constructor. can be null.
     * @param parameters the parameter values to the user defined constructor. can be null.
     *
     * @return an instance created by calling the constructor this className represents.
     *
     * @throws PersistenceConfigurationException if any error occurs in creating the new instance for given parameters
     */
    public static Object createObject(Class classType, String className,
        Class[] types, Object[] parameters)
        throws PersistenceConfigurationException {
        try {
            Object object = null;

            // types is null, use default constructor
            if (types == null) {
                object = Class.forName(className).newInstance();
            } else {
                object = Class.forName(className).getConstructor(types)
                              .newInstance(parameters);
            }

            // Created Object is not instance of given classType
            if (!(classType.isInstance(object))) {
                throw new PersistenceConfigurationException("Class "
                        + className + " does not implement " + classType.getName());
            }

            return object;
        } catch (ClassNotFoundException e) {
            throw new PersistenceConfigurationException(
                "Unable to locate the class " + className, e);
        } catch (NoSuchMethodException e) {
            throw new PersistenceConfigurationException("Unable to locate the constructor",
                e);
        } catch (InstantiationException e) {
            throw new PersistenceConfigurationException("Fails to create the instance",
                e);
        } catch (IllegalAccessException e) {
            throw new PersistenceConfigurationException("Fails to create the instance",
                e);
        } catch (InvocationTargetException e) {
            throw new PersistenceConfigurationException("Fails to create the instance",
                e);
        }
    }

    /**
     * <p>
     * Serializes the given object into a byte array using Java Serialization API.
     * </p>
     *
     * <p>
     * It is used when we want to save an object to persistence(for example, XML file, database).
     * </p>
     *
     * @param obj the given object to serialize
     * @return the serialized byte array
     * @throws IndexPersistenceException when IOException occurs during the serialization process.
     */
    public static byte[] serializeObject(Object obj)
        throws IndexPersistenceException {
        try {
            // standard java serialization mechanism is used
            ByteArrayOutputStream byteArrayOut = new ByteArrayOutputStream();
            ObjectOutputStream out = new ObjectOutputStream(byteArrayOut);
            out.writeObject(obj);

            return byteArrayOut.toByteArray();
        } catch (IOException ioe) {
            throw new IndexPersistenceException("IOException occurs when serializing object.",
                ioe);
        }
    }

    /**
     * <p>
     * Deserializes the given byte array into an object using Java Serialization API.
     * </p>
     *
     * <p>
     * It is used when we want to recover an object from persistence(for example, XML file, database).
     * </p>
     *
     * @param data the byte array represents an object
     * @return the object deserialized from the given byte array
     * @throws IndexPersistenceException when IOException or ClassNotFoundException occurs during the
     * deserialization process.
     */
    public static Object deSerializeObject(byte[] data)
        throws IndexPersistenceException {
        try {
            // standard java serialization mechanism is used
            InputStream byteArrayIn = new ByteArrayInputStream(data);
            ObjectInputStream in = new ObjectInputStream(byteArrayIn);

            return in.readObject();
        } catch (IOException ioe) {
            throw new IndexPersistenceException("IOException occurs when deserializing object.",
                ioe);
        } catch (ClassNotFoundException e) {
            throw new IndexPersistenceException("ClassNotFoundException occurs when deserializing object.",
                e);
        }
    }

    /**
     * <p>
     * Returns the unique string id for the given WordSourceId instance.
     * </p>
     *
     * @param wordSourceId the given WordSourceId instance to get its unique string id representation.
     * @return the unique string id for the given WordSourceId instance
     * @throws IndexPersistenceException when the MD5 algorithm is not supported
     */
    public static String getDigestBasedId(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        try {
            // Create an MD5 message digest
            MessageDigest md = MessageDigest.getInstance("MD5");

            // serialize WordSourceId instance
            byte[] bytes = getWordSourceIdBytes(wordSourceId);

            return byte2hexString(md.digest(bytes));

        } catch (NoSuchAlgorithmException e) {
            throw new IndexPersistenceException("NoSuchAlgorithmException occurs when generating document index id.",
                e);
        }
    }

    /**
     * <p>
     * Encodes the given byte array into a string using BASE64 algorithm.
     * </p>
     *
     * <p>
     * It is used when we want to save a byte stream to text-base persistence(for example, XML file).
     * The main usage of this method is to avoid control characters and allow only displayable characters.
     * </p>
     *
     * @param rawObjectBytes the given byte array to encode
     * @return the string encoded by BASE 64 algorithm
     * @throws IndexPersistenceException when UTF-8 encoding is not supported, which should never happen.
     */
    public static final String base64Encode(byte[] rawObjectBytes)
        throws IndexPersistenceException {
        // create an encoder
        // we just have one long line of output using the standard alphabet
        final Base64Encoder encoder = new Base64Encoder(0, null, Base64Codec.STANDARD_ALPHABET);

        // set encode input
        encoder.setInput(rawObjectBytes);
        encoder.finish();

        // get encoded bytes
        //We create a buffer to hold the encoded data. Since we do
        //not know the exact size but we know that roughly 1.3
        //times will be needed we allocate 2*original string just in case.
        byte[] encodedObjectBytes = new byte[rawObjectBytes.length * 2];
        int len = encoder.deflate(encodedObjectBytes);
        encoder.end();

        // now the 0..len bytes hold our encoded data to get the String representation in UTF-8 encoding.
        try {
            return new String(encodedObjectBytes, 0, len, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new IndexPersistenceException("UnsupportedEncodingException occurs when encoding bytes.",
                e);
        }
    }

    /**
     * <p>
     * Decodes the given string into a byte array using BASE64 algorithm.
     * </p>
     *
     * <p>
     * It is used when we want to recover a byte stream from text-base persistence(for example, XML file).
     * Here, BASE64 algorithm is used and the given data must be encoded using BASE64 algorithm.
     * </p>
     *
     *
     * @param data the given string to convert to a byte array using base64 algorithm
     * @return the converted byte array
     * @throws IndexPersistenceException when DataFormatException or UnsupportedEncodingException occurs.
     */
    public static final byte[] base64Decode(String data)
        throws IndexPersistenceException {
        // create a BASE 64 decoder
        final Base64Decoder decoder = new Base64Decoder(Base64Codec.IGNORE_NONE,
                null, Base64Codec.STANDARD_ALPHABET);

        try {
            byte[] bytes = data.getBytes("UTF-8");

            // set decode input
            decoder.setInput(bytes);

            // get encoded bytes
            byte[] bufferedBytes = new byte[bytes.length * 2];
            int len = decoder.inflate(bufferedBytes);

            byte[] decodedObjectBytes = new byte[len];
            System.arraycopy(bufferedBytes, 0, decodedObjectBytes, 0, len);

            return decodedObjectBytes;
        } catch (DataFormatException e) {
            throw new IndexPersistenceException("DataFormatException occurs when decoding data.",
                e);
        } catch (UnsupportedEncodingException e) {
            throw new IndexPersistenceException("UnsupportedEncodingException occurs when decoding data.",
                e);
        }
    }

    /**
     * <p>
     * Encodes the given string into a string using Hex code, that is, every char in the string is converted
     * to an integer and then saved as a hex string.
     * </p>
     *
     * <p>
     * It is used when we want to save string which contains control characters to text-base persistence
     * (for example, XML file).
     * Specially, this method is used to encode delimiters.
     * </p>
     *
     * @param data the given string to encode
     * @return the encoded string using hex code
     */
    public static String hexCode(String data) {
        StringBuffer buffer = new StringBuffer();

        // convert every character to its hex string representation
        for (int j = 0; j < data.length(); j++) {
            char c = data.charAt(j);
            buffer.append(new StringBuffer("0x"));
            buffer.append(Integer.toHexString((int) c));
        }

        return buffer.toString();
    }

    /**
     * <p>
     * Decodes the given string into a string which is encoded in Hex, that is, every integer representation
     * in the string is parsed as a hex string and then converted to an char and.
     * </p>
     *
     * <p>
     * It is used when we want to recover the original string which contains control characters from text-base
     * persistence (for example, XML file).
     * Specially, this method is used to decode delimiters.
     * </p>
     *
     * @param data the given string to decode
     * @return the original string before encoding
     */
    public static String hexDecode(String data) {
        StringBuffer buffer = new StringBuffer();
        String[] values = data.split("0x");

        // recover every character from its hex string representation
        for (int i = 0; i < values.length; i++) {
            if ((values[i] == null) || (values[i].trim().length() == 0)) {
                continue;
            }

            buffer.append((char) Integer.parseInt(values[i], HEX));
        }

        return buffer.toString();
    }

    /**
     * <p>
     * Converts a WordSourceId instance to a byte array using Java Serialization API.
     * </p>
     *
     * <p>
     * Note, the direct serialization api cannot be used because WordSourceId is not declared as Serializable.
     * However, all the fields in WordSourceId are serializable, so all the fields will be serialization and then
     * joined them together.
     * </p>
     *
     * @param wordSourceId the given WordSourceId instance to convert
     * @return the byte array representation of the given WordSourceId instance
     * @throws IndexPersistenceException if any field of the given WordSourceId is not serializable.
     */
    private static byte[] getWordSourceIdBytes(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        try {
            // serialize all the fields in WordSourceId instance
            // and join them together
            ByteArrayOutputStream byteArrayOut = new ByteArrayOutputStream();
            byteArrayOut.write(serializeObject(wordSourceId.getDelimiters()));
            byteArrayOut.write(serializeObject(wordSourceId.getSourceIdentity()));
            byteArrayOut.write(serializeObject(
                    wordSourceId.getSourceClassName()));
            byteArrayOut.write(serializeObject(wordSourceId.getSourceLocale()));

            return byteArrayOut.toByteArray();
        } catch (IOException ioe) {
            throw new IndexPersistenceException("IOException occurs when serializing word source id.",
                ioe);
        }
    }

    /**
     * <p>
     * Gets the string representation of the given byte array using hex representations of every byte in byte array.
     * </p>
     *
     * <p>
     * Note, this method is only used by getDigestBasedId() method.
     * </p>
     *
     * @param data the byte array to convert
     * @return the string representation of the given byte array using hex representations of every byte in byte array.
     */
    private static String byte2hexString(byte[] data) {
        StringBuffer buf = new StringBuffer(data.length * 2);

        // convert every byte to hex string
        for (int i = 0; i < data.length; i++) {
            if (((int) data[i] & CONSTANT) < HEX) {
                buf.append("0");
            }

            buf.append(Long.toString((int) data[i] & CONSTANT, HEX));
        }

        return buf.toString();
    }
}
