/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.ConfigurationException;
import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;
import com.topcoder.db.connectionfactory.UnknownConnectionException;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.DocumentIndexPersistenceUtil;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import java.text.CollationKey;
import java.text.Collator;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;


/**
 * <p>
 * This class is used as a utility class for the common database operation.
 * </p>
 *
 * <p>
 * In this component, simple access solution and fast access solution are provided, they are mostly similar in table
 * schemas, and so all common operations related to the database are going to this class for code reused.
 * </p>
 *
 * <p>
 * Note, the common operations are reused based on the table schemas, If, in the future, the table schemas change
 * and no longer share something in common, then the future developer will be responsible for writing new code to
 * deal with the situation.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
final class DBUtil {
    /**
     * <p>
     * Represents the select all document sql string constant.
     * </p>
     */
    private static final String SELECT_ALL_DOCUMENT = "SELECT DOCUMENT_ID FROM DOCUMENT";

    /**
     * <p>
     * Represents the select document sql string constant, which is based on document id.
     * </p>
     */
    private static final String SELECT_DOCUMENT = "SELECT LOCALE, SOURCE_IDENTITY, CLASS_NAME, USE_COUNT FROM DOCUMENT "
        + "WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the select word sql string constant, which is based on document id.
     * </p>
     */
    private static final String SELECT_WORD = "SELECT WORD_ID, WORD_ENTRY FROM WORD WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the select word position sql string constant, which is based on document id.
     * </p>
     */
    private static final String SELECT_WORD_POSITION = "SELECT WORD_ID, WORD_POSITION FROM WORD_POSITION "
        + "WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the select word data sql string constant, which is based on document id.
     * </p>
     */
    private static final String SELECT_WORD_DATA = "SELECT WORD_DATA FROM WORD_DATA WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the select delimiter sql string constant, which is based on document id.
     * </p>
     */
    private static final String SELECT_DELIMITER = "SELECT DELIMITER_ENTRY FROM DELIMITER WHERE DOCUMENT_ID = ?"
        + " ORDER BY DELIMITER_SORT_ORDER";

    /**
     * <p>
     * Represents the select document id that related to the given collection id sql string constant.
     * </p>
     */
    private static final String SELECT_DOCUMENT_COLLECTION_DOCUMENT_XREF = "SELECT DOCUMENT_ID FROM "
        + "DOCUMENT_COLLECTION_DOCUMENT_XREF WHERE COLLECTION_ID = ?";

    /**
     * <p>
     * Represents the delete document sql string constant, which is based on document id.
     * </p>
     */
    private static final String DELETE_DOCUMENT = "DELETE FROM DOCUMENT WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the delete delimiter sql string constant, which is based on document id.
     * </p>
     */
    private static final String DELETE_DELIMITER = "DELETE FROM DELIMITER WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the delete word sql string constant, which is based on document id.
     * </p>
     */
    private static final String DELETE_WORD = "DELETE FROM WORD WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the delete word position sql string constant, which is based on document id.
     * </p>
     */
    private static final String DELETE_WORD_POSITION = "DELETE FROM WORD_POSITION WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the delete word data sql string constant, which is based on document id.
     * </p>
     */
    private static final String DELETE_WORD_DATA = "DELETE FROM WORD_DATA WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the delete the relationship between document id and collection id sql string constant.
     * </p>
     */
    private static final String DELETE_DOCUMENT_COLLECTION_DOCUMENT_XREF =
        "DELETE FROM DOCUMENT_COLLECTION_DOCUMENT_XREF"
        + " WHERE COLLECTION_ID =? AND DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the delete all the relationships for the given collection id sql string constant.
     * </p>
     */
    private static final String DELETE_DCDX_BY_COLLECTION_ID = "DELETE FROM DOCUMENT_COLLECTION_DOCUMENT_XREF"
        + " WHERE COLLECTION_ID = ?";

    /**
     * <p>
     * Represents the delete collection sql string constant, which is based on collection id.
     * </p>
     */
    private static final String DELETE_DOCUMENT_COLLECTION = "DELETE FROM DOCUMENT_COLLECTION WHERE COLLECTION_ID = ?";

    /**
     * <p>
     * Represents the insert document sql string constant.
     * </p>
     */
    private static final String INSERT_DOCUMENT = "INSERT INTO DOCUMENT (DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, "
        + "CLASS_NAME, USE_COUNT) VALUES (?, ?, ?, ?, ?)";

    /**
     * <p>
     * Represents the insert delimiter sql string constant.
     * </p>
     */
    private static final String INSERT_DELIMITER = "INSERT INTO DELIMITER (DOCUMENT_ID, DELIMITER_SORT_ORDER, "
        + "DELIMITER_ENTRY) VALUES (?, ?, ?)";

    /**
     * <p>
     * Represents the insert word sql string constant.
     * </p>
     */
    private static final String INSERT_WORD = "INSERT INTO WORD (DOCUMENT_ID, WORD_ID, WORD_ENTRY) VALUES (?, ?, ?)";

    /**
     * <p>
     * Represents the insert word data sql string constant.
     * </p>
     */
    private static final String INSERT_WORD_DATA = "INSERT INTO WORD_DATA (DOCUMENT_ID, WORD_DATA) VALUES (?, ?)";

    /**
     * <p>
     * Represents the insert word position sql string constant.
     * </p>
     */
    private static final String INSERT_WORD_POSITION = "INSERT INTO WORD_POSITION (DOCUMENT_ID, WORD_ID, WORD_POSITION)"
        + " VALUES (?, ?, ?)";

    /**
     * <p>
     * Represents the insert document collection sql string constant.
     * </p>
     */
    private static final String INSERT_DOCUMENT_COLLECTION = "INSERT INTO DOCUMENT_COLLECTION (COLLECTION_ID) VALUES "
        + "(?)";

    /**
     * <p>
     * Represents the insert relationship between document and collection document sql string constant.
     * </p>
     */
    private static final String INSERT_DOCUMENT_COLLECTION_DOCUMENT_XREF = "INSERT INTO "
        + "DOCUMENT_COLLECTION_DOCUMENT_XREF (COLLECTION_ID, DOCUMENT_ID) VALUES (?, ?)";

    /**
     * <p>
     * Represents the select document collection sql string constant.
     * </p>
     */
    private static final String SELECT_DOCUMENT_COLLECTION = "SELECT * FROM DOCUMENT_COLLECTION WHERE "
        + "COLLECTION_ID = ?";

    /**
     * <p>
     * Represents the select update document sql string constant.
     * </p>
     */
    private static final String UPDATE_DOCUMENT = "UPDATE DOCUMENT SET USE_COUNT = ? WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * Represents the DOCUMENT_ID column constant.
     * </p>
     */
    private static final String DOCUMENT_ID = "DOCUMENT_ID";

    /**
     * <p>
     * Represents the LOCALE column constant.
     * </p>
     */
    private static final String LOCALE = "LOCALE";

    /**
     * <p>
     * Represents the SOURCE_IDENTITY column constant.
     * </p>
     */
    private static final String SOURCE_IDENTITY = "SOURCE_IDENTITY";

    /**
     * <p>
     * Represents the USE_COUNT column constant.
     * </p>
     */
    private static final String USE_COUNT = "USE_COUNT";

    /**
     * <p>
     * Represents the CLASS_NAME column constant.
     * </p>
     */
    private static final String CLASS_NAME = "CLASS_NAME";

    /**
     * <p>
     * Represents the WORD_POSITION column constant.
     * </p>
     */
    private static final String WORD_POSITION = "WORD_POSITION";

    /**
     * <p>
     * Represents the DELIMITER_ENTRY column constant.
     * </p>
     */
    private static final String DELIMITER_ENTRY = "DELIMITER_ENTRY";

    /**
     * <p>
     * Represents the WORD_ID column constant.
     * </p>
     */
    private static final String WORD_ID = "WORD_ID";

    /**
     * <p>
     * Represents the WORD_DATA column constant.
     * </p>
     */
    private static final String WORD_DATA = "WORD_DATA";

    /**
     * <p>
     * Represents the WORD_ENTRY column constant.
     * </p>
     */
    private static final String WORD_ENTRY = "WORD_ENTRY";

    /**
     * <p>
     * Represent the connectionFactoryClass property name.
     * </p>
     */
    private static final String CONNECTION_FACTORY_CLASS = "connectionFactoryClassName";

    /**
     * <p>
     * Represent the connectionFactoryNamespace property name.
     * </p>
     */
    private static final String CONNECTION_FACTORY_NAMESPACE = "connectionFactoryNamespace";

    /**
     * <p>
     * This private constructor prevents to create a new instance.
     * </p>
     */
    private DBUtil() {
    }

    /**
     * <p>
     * Saves the document and delimiters of the given DocumentIndex instance into database.
     * </p>
     *
     * <p>
     * Note, this database operation relates two tables : DOCUMENT and DELIMITER.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndex a DocumentIndex instance for inserting.
     * @param documentIndexId the unique string id of the given DocumentIndex instance.
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when exception occurs during the serialization process.
     */
    static void insertDocument(Connection conn, DocumentIndex documentIndex,
        String documentIndexId) throws SQLException, IndexPersistenceException {
        // insert record into table DOCUMENT
        List documentArgs = new LinkedList();
        documentArgs.add(documentIndexId);
        documentArgs.add(documentIndex.getLocale().toString());

        // source identity object will be deserialized before inserting
        Object sourceIdentity = documentIndex.getWordSourceId().getSourceIdentity();
        byte[] data = DocumentIndexPersistenceUtil.serializeObject(sourceIdentity);

        documentArgs.add(data);
        documentArgs.add(documentIndex.getWordSourceId().getSourceClassName());
        documentArgs.add(new Integer(0));
        executeUpdate(conn, INSERT_DOCUMENT, documentArgs);


        // insert records into table DELIMITER
        String[] delimeters = documentIndex.getWordSourceId().getDelimiters();

        for (int i = 0; i < delimeters.length; i++) {
            List delimeterArgs = new LinkedList();
            delimeterArgs.add(documentIndexId);
            delimeterArgs.add(new Integer(i));
            delimeterArgs.add(delimeters[i]);
            executeUpdate(conn, INSERT_DELIMITER, delimeterArgs);
        }
    }

    /**
     * <p>
     * Saves the words information of the DocumentIndex instance into database.
     * </p>
     *
     * <p>
     * This operation is only executed by the simple access solution.
     * </p>
     *
     * <p>
     * Note, this database operation relates two tables : WORD and WORD_POSITION.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndex a DocumentIndex instance for inserting.
     * @param documentIndexId the unique string id of the given DocumentIndex instance.
     * @throws SQLException when exception occurs during the database operation
     */
    static void simpleInsertWords(Connection conn, DocumentIndex documentIndex,
        String documentIndexId) throws SQLException {
        Map words = documentIndex.getWords();

        int wordsSize = 0;

        for (Iterator it = words.entrySet().iterator(); it.hasNext();) {
            wordsSize++;

            // insert word record to database
            List wordsArgs = new LinkedList();
            Entry entry = (Entry) it.next();
            CollationKey word = (CollationKey) entry.getKey();
            List wordPosList = (List) entry.getValue();
            Integer curWordId = new Integer(wordsSize);
            wordsArgs.add(documentIndexId);
            wordsArgs.add(curWordId);
            wordsArgs.add(word.getSourceString());
            executeUpdate(conn, INSERT_WORD, wordsArgs);

            for (Iterator pIt = wordPosList.iterator(); pIt.hasNext();) {
                List wordPositionsArgs = new LinkedList();
                wordPositionsArgs.add(documentIndexId);
                wordPositionsArgs.add(curWordId);
                wordPositionsArgs.add(pIt.next());
                // insert word position record to database
                executeUpdate(conn, INSERT_WORD_POSITION, wordPositionsArgs);
            }
        }
    }

    /**
     * <p>
     * Saves the words information of the DocumentIndex instance into database.
     * </p>
     *
     * <p>
     * This operation is only executed by the fast access solution.
     * </p>
     *
     * <p>
     * Note, this database operation relates one table : WORD_DATA.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndex a DocumentIndex instance for inserting.
     * @param documentIndexId the unique string id of the given DocumentIndex instance.
     * @throws SQLException when exception occurs during the database operation
     */
    static void fastInsertWords(Connection conn, DocumentIndex documentIndex,
        String documentIndexId) throws SQLException {
        Map words = documentIndex.getWords();

        // we will use first delimiter, it's guaranteed to be not empty
        String delimiter = documentIndex.getWordSourceId().getDelimiters()[0];

        StringBuffer buffer = new StringBuffer();

        for (Iterator it = words.entrySet().iterator(); it.hasNext();) {
            Entry entry = (Entry) it.next();
            CollationKey word = (CollationKey) entry.getKey();
            List wordPosList = (List) entry.getValue();
            buffer.append(word.getSourceString());

            for (Iterator pIt = wordPosList.iterator(); pIt.hasNext();) {
                // use single delimiter to separate different word positions
                buffer.append(delimiter);
                buffer.append(pIt.next());
            }

            // use double delimiter to separate different words
            if (it.hasNext()) {
                buffer.append(delimiter);
                buffer.append(delimiter);
            }
        }

        List wordsArgs = new LinkedList();
        wordsArgs.add(documentIndexId);
        wordsArgs.add(buffer.toString().getBytes());

        // insert all the words in a single record in database
        executeUpdate(conn, INSERT_WORD_DATA, wordsArgs);
    }

    /**
     * <p>
     * Queries the words information from database and returned as a Map instance.
     * In the returned Map instance, key is instance of CollationKey and value is instance of List
     * (a list contains position integer).
     * </p>
     *
     * <p>
     * This operation is only executed by the simple access solution.
     * </p>
     *
     * <p>
     * Note, this database operation relates two tables : WORD and WORD_POSITION.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndexId the unique string id of the given WordSourceId instance.
     * @param wordSourceId the WordSourceId instance for querying
     * @return a words Map, in which key is collation key and value is position list
     * @throws SQLException when exception occurs during the database operation
     */
    static Map simpleQueryWords(Connection conn, String documentIndexId,
        WordSourceId wordSourceId) throws SQLException {
        Map words = new HashMap();
        List args = Collections.singletonList(documentIndexId);
        List wordList = DBUtil.executeQuery(conn, SELECT_WORD, args);
        Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());
        Map entries = new HashMap();

        // select all the words from database
        // the position list will be initialized empty.
        for (Iterator it = wordList.iterator(); it.hasNext();) {
            Map map = (Map) it.next();
            Integer wordId = (Integer) map.get(WORD_ID);
            String word = (String) map.get(WORD_ENTRY);
            CollationKey key = collator.getCollationKey(word);
            words.put(key, new ArrayList());
            entries.put(wordId, key);
        }

        List positionList = DBUtil.executeQuery(conn, SELECT_WORD_POSITION, args);

        // select all the positions related to the document index id
        // the corresponding word must exist because of database constraints.
        for (Iterator it = positionList.iterator(); it.hasNext();) {
            Map map = (Map) it.next();
            Integer wordId = (Integer) map.get(WORD_ID);
            Integer pos = (Integer) map.get(WORD_POSITION);
            CollationKey key = (CollationKey) entries.get(wordId);
            List wordPositions = (List) words.get(key);
            wordPositions.add(pos);
        }

        return words;
    }

    /**
     * <p>
     * Queries the words information from database and returned as a Map instance.
     * In the returned Map instance, key is instance of CollationKey and value is instance of List
     * (a list contains position integer).
     * </p>
     *
     * <p>
     * This operation is only executed by the fast access solution.
     * </p>
     *
     * <p>
     * Note, this database operation relates one table : WORD_DATA.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndexId the unique string id of the given WordSourceId instance.
     * @param wordSourceId the WordSourceId instance for querying
     * @return a words Map, in which key is collation key and value is position list
     * @throws SQLException when exception occurs during the database operation
     */
    static Map fastQueryWords(Connection conn, String documentIndexId,
        WordSourceId wordSourceId) throws SQLException {
        List wordList = DBUtil.executeQuery(conn, SELECT_WORD_DATA,
                Collections.singletonList(documentIndexId));
        Map wordsMap = new HashMap();
        Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());

        if (wordList.size() != 0) {
            Map map = (Map) wordList.get(0);

            // all the words are saved as a CLOB column
            Clob clob = (Clob) map.get(WORD_DATA);
            String data = clob.getSubString(1, (int) clob.length());

            // first delimiter will be used
            String delimiter = wordSourceId.getDelimiters()[0];

            // separate different words
            String[] words = data.split(delimiter + delimiter);

            for (int i = 0; i < words.length; i++) {

                // separate different positions for each word
                String[] singleWordData = words[i].split(delimiter);

                for (int j = 0; j < singleWordData.length; j++) {
                    String word = singleWordData[0];
                    List wordPositions = new ArrayList();

                    for (int k = 1; k < singleWordData.length; k++) {
                        wordPositions.add(Integer.valueOf(singleWordData[k]));
                    }

                    CollationKey key = collator.getCollationKey(word);
                    wordsMap.put(key, wordPositions);
                }
            }
        }

        return wordsMap;
    }

    /**
     * <p>
     * Deletes all the data related to a given document index id.
     * </p>
     *
     * <p>
     * The operation can be called by either simple access solution or fast access solution.
     * When isSimple flag is true, then simple access access solution otherwise fast access solution.
     * </p>
     *
     * <p>
     * Note, this database operation relates five tables : WORD_POSITION, WORD, WORD_DATA,
     * DELIMITER and DOCUMENT.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndexId the unique string id of WordSourceId instance.
     * @param isSimple a flag to identify whether it is called by simple access solution or fast access solution.
     * When it's true, then simple access access solution otherwise fast access solution.
     * @throws SQLException when exception occurs during the database operation
     */
    static void removeDocumentIndex(Connection conn, String documentIndexId,
        boolean isSimple) throws SQLException {
        List args = Collections.singletonList(documentIndexId);

        if (isSimple) {
            // this is for simple access solution
            executeUpdate(conn, DELETE_WORD_POSITION, args);
            executeUpdate(conn, DELETE_WORD, args);
        } else {
            // this is for fast access solution
            executeUpdate(conn, DELETE_WORD_DATA, args);
        }

        // remove delimiters and document records
        executeUpdate(conn, DELETE_DELIMITER, args);
        executeUpdate(conn, DELETE_DOCUMENT, args);
    }

    /**
     * <p>
     * Queries the database to return a DocumentIndex instance based on the given documentIndexId.
     * The document, delimiter, words are all queried and wraps as a DocumentIndex instance.
     * </p>
     *
     * <p>
     * The operation can be called by either simple access solution or fast access solution.
     * When isSimple flag is true, then simple access access solution otherwise fast access solution.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndexId the unique string id of WordSourceId instance.
     * @param isSimple a flag to identify whether it is called by simple access solution or fast access solution.
     * When it's true, then simple access access solution otherwise fast access solution.
     * @return a DocumentIndex instance based on the given documentIndexId.
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when fails to deserialize the source identity object
     */
    static DocumentIndex getDocumentIndex(Connection conn,
        String documentIndexId, boolean isSimple)
        throws SQLException, IndexPersistenceException {

        // first query tables to create the new WordSourceId instance
        WordSourceId wordSourceId = getWordSourceId(conn, documentIndexId);

        Map words = null;

        // then query words information depends on simple access solution or fast access solution
        if (isSimple) {
            words = simpleQueryWords(conn, documentIndexId, wordSourceId);
        } else {
            words = fastQueryWords(conn, documentIndexId, wordSourceId);
        }

        return new DocumentIndex(wordSourceId, words);
    }

    /**
     * <p>
     * Queries the database to return a WordSourceId instance based on the given documentIndexId.
     * The document, delimiter are all queried and wraps as a WordSourceId instance.
     * </p>
     *
     * <p>
     * Note, this operation relates to tables DOCUMENT and DELIMITER.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndexId the unique string id of WordSourceId instance.
     * @return a WordSourceId instance based on the given documentIndexId.
     * @throws IndexPersistenceException when fails to deserialize the source identity object
     * @throws SQLException when exception occurs during the database operation
     */
    static WordSourceId getWordSourceId(Connection conn, String documentIndexId)
        throws IndexPersistenceException, SQLException {
        // query the table DOCUMENT
        Document document = queryDocument(conn, documentIndexId);

        // query the table DELIMITER
        List delimiterList = DBUtil.executeQuery(conn, SELECT_DELIMITER,
                Collections.singletonList(documentIndexId));
        String[] delimiters = new String[delimiterList.size()];

        for (int i = 0; i < delimiterList.size(); i++) {
            Map delimitiersMap = (Map) delimiterList.get(i);
            delimiters[i] = (String) delimitiersMap.get(DELIMITER_ENTRY);
        }

        return new WordSourceId(document.getSourceIdentity(),
            document.getClassName(), delimiters, document.getLocale());
    }

    /**
     * <p>
     * Queries the database to return a list instance that contains all the document index ids which are related to
     * the given collection id.
     * The element in the returned list is document index id, which is type of String.
     * </p>
     *
     * <p>
     * Note, this operation relates to table DOCUMENT_COLLECTION_DOCUMENT_XREF.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param collectionId the unique string id of collection document
     * @return a list instance that contains all the document index ids which are related to the given collection id.
     * The element in the returned list is document index id, which is type of String.
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when fails to deserialize the source identity object
     */
    static List getDocuments(Connection conn, String collectionId)
        throws SQLException, IndexPersistenceException {

        // query the relationships from database DOCUMENT_COLLECTION_DOCUMENT_XREF
        List result = DBUtil.executeQuery(conn,
                SELECT_DOCUMENT_COLLECTION_DOCUMENT_XREF,
                Collections.singletonList(collectionId));
        List documents = new ArrayList();

        for (Iterator it = result.iterator(); it.hasNext();) {
            Map map = (Map) it.next();
            String documentIndexId = (String) map.get(DOCUMENT_ID);
            documents.add(queryDocument(conn, documentIndexId));
        }

        return documents;
    }

    /**
     * <p>
     * Returns the CollectionIndex instance from database based on the given document collection id.
     * Note, all the documents associated with the CollectionIndex id will be retrieved as well.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param collectionId the unique string id of collection document
     * @param isSimple a flag to identify whether it is called by simple access solution or fast access solution.
     * When it's true, then simple access access solution otherwise fast access solution.
     * @return the CollectionIndex instance from database based on the given document collection id.
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when fails to load CollectionIndex instance.
     */
    static CollectionIndex getCollectionIndex(Connection conn,
        String collectionId, boolean isSimple)
        throws SQLException, IndexPersistenceException {
        Map wordsMap = new HashMap();
        Set wordSourceIds = new HashSet();

        // select all the related document index ids
        List result = DBUtil.executeQuery(conn,
                SELECT_DOCUMENT_COLLECTION_DOCUMENT_XREF,
                Collections.singletonList(collectionId));

        for (Iterator it = result.iterator(); it.hasNext();) {
            Map map = (Map) it.next();
            String documentIndexId = (String) map.get(DOCUMENT_ID);

            // query complete information for document index
            DocumentIndex documentIndex = getDocumentIndex(conn,
                    documentIndexId, isSimple);

            WordSourceId wordSourceId = documentIndex.getWordSourceId();
            wordSourceIds.add(wordSourceId);

            // converts the words in document index to be suitable for document collection
            for (Iterator iter = documentIndex.getWords().keySet().iterator();
                    iter.hasNext();) {
                CollationKey key = (CollationKey) iter.next();
                Set documentIds = (Set) wordsMap.get(key);

                if (documentIds == null) {
                    // The word does not exist, create a new list
                    documentIds = new HashSet();
                    wordsMap.put(key, documentIds);
                }

                // Mark the word's existence in the document
                documentIds.add(wordSourceId);
            }
        }

        return new CollectionIndex(null, wordsMap, wordSourceIds, collectionId);
    }

    /**
     * <p>
     * Queries the database whether the given collection id exist.
     * </p>
     *
     * <p>
     * Note, this operation relates to table DOCUMENT_COLLECTION.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param collectionId the unique string id of collection document
     * @return true if the given collection record found otherwise return false
     * @throws SQLException when exception occurs during the database operation
     */
    static boolean isCollectionIndexPersisted(Connection conn,
        String collectionId) throws SQLException {
        // query the table DOCUMENT
        List result = DBUtil.executeQuery(conn, SELECT_DOCUMENT_COLLECTION,
                Collections.singletonList(collectionId));

        return (result.size() != 0);
    }

    /**
     * <p>
     * Removes the collection record from database, which includes deleting all the relationships for the given
     * collection id,
     * updating the corresponding document usage and deleting the collection record itself.
     * </p>
     *
     * <p>
     * Note, this operation relates to tables DOCUMENT_COLLECTION, DOCUMENT and DOCUMENT_COLLECTION_DOCUMENT_XREF.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param collectionId the unique string id of collection document
     * @param documents a list contains all the document index ids that relate to the given collection id.
     * @throws SQLException when exception occurs during the database operation
     */
    static void removeCollectionIndex(Connection conn, String collectionId,
        List documents) throws SQLException {
        List args = Collections.singletonList(collectionId);
        DBUtil.executeUpdate(conn, DELETE_DCDX_BY_COLLECTION_ID, args);

        for (Iterator it = documents.iterator(); it.hasNext();) {
            Document doc = (Document) it.next();
            // use count will be updated
            updateDocumentUsageCount(conn, doc.getDocumentId(), doc.getUseCount() - 1);
        }

        DBUtil.executeUpdate(conn, DELETE_DOCUMENT_COLLECTION, args);
    }

    /**
     * <p>
     * Adds a collection record into database.
     * </p>
     *
     * <p>
     * This operation includes inserting the collection record itself, inserting the relationships for the
     * given collection record and updating the corresponding document usage.
     * </p>
     *
     * <p>
     * Note, this operation relates to tables DOCUMENT_COLLECTION, DOCUMENT, DELIMITER and
     * DOCUMENT_COLLECTION_DOCUMENT_XREF.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param collectionIndex ColectionIndex instance
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when fails to get unique string id for a WordSourceId instance or
     * deserialize the source identity object
     */
    static void insertCollectionIndex(Connection conn,
        CollectionIndex collectionIndex)
        throws SQLException, IndexPersistenceException {
        List args = Collections.singletonList(collectionIndex.getId());
        DBUtil.executeUpdate(conn, INSERT_DOCUMENT_COLLECTION, args);

        List xrefArgs = new LinkedList();
        xrefArgs.add(collectionIndex.getId());

        for (Iterator it = collectionIndex.getAllDocumentIds().iterator();
                it.hasNext();) {
            WordSourceId wordSourceId = (WordSourceId) it.next();
            String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);
            Document currentDocument = queryDocument(conn, documentIndexId);
            xrefArgs.add(documentIndexId);

            // update relationships
            executeUpdate(conn, INSERT_DOCUMENT_COLLECTION_DOCUMENT_XREF, xrefArgs);

            // use count will be updated
            updateDocumentUsageCount(conn, documentIndexId,
                currentDocument.getUseCount() + 1);
            xrefArgs.remove(xrefArgs.size() - 1);
        }
    }

    /**
     * <p>
     * Queries the database to return a Document instance, which includes information stored in table
     * DOCUMENT and DELIMITER.
     * </p>
     *
     * <p>
     * Note, this operation relates two tables : DOCUMENT and DELIMITER.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndexId the unique string id of WordSourceId instance
     * @return a Document instance based on the given document index id
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when fails to deserialize the source identity object
     */
    static Document queryDocument(Connection conn, String documentIndexId)
        throws SQLException, IndexPersistenceException {
        List args = Collections.singletonList(documentIndexId);
        List result = DBUtil.executeQuery(conn, SELECT_DOCUMENT, args);

        // if the document index id doesn't exist in database, null is expected.
        if (result.size() == 0) {
            return null;
        } else {
            Map map = (Map) result.get(0);
            String localeStr = (String) map.get(LOCALE);

            // decode the source identity object, serialization technology will be used
            Blob blob = ((Blob) map.get(SOURCE_IDENTITY));
            int length = (int) blob.length();
            byte[] data = blob.getBytes(1, length);
            Object sourceIdentity = DocumentIndexPersistenceUtil.deSerializeObject(data);

            String className = (String) map.get(CLASS_NAME);
            int useCount = ((Integer) map.get(USE_COUNT)).intValue();

            // recovers a Locale instance from locale string(the string is from Locale.toString())
            String[] localeParts = localeStr.trim().split("_");
            Locale locale = null;

            if (localeParts.length == 1) {
                locale = new Locale(localeParts[0]);
            } else if (localeParts.length == 2) {
                locale = new Locale(localeParts[0], localeParts[1]);
            } else {
                locale = new Locale(localeParts[0], localeParts[1],
                        localeParts[2]);
            }

            return new Document(documentIndexId, locale, sourceIdentity,
                className, useCount);
        }
    }

    /**
     * <p>
     * Queries the database to return a set instance which contains WordSourceId instance.
     * </p>
     *
     * <p>
     * This operation will return all the document index records in database and returned in WordSourceId instance.
     * </p>
     *
     * <p>
     * Note, this operation relates to tables DOCUMENT and DELIMITER.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @return a set instance which contains all WordSourceId instances found in database.
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when fails to load all the indexed documents.
     */
    static Set getIndexedDocuments(Connection conn)
        throws SQLException, IndexPersistenceException {
        Set wordSourceIds = new HashSet();

        // query all the document records
        List result = executeQuery(conn, SELECT_ALL_DOCUMENT,
                Collections.EMPTY_LIST);

        for (Iterator it = result.iterator(); it.hasNext();) {
            Map map = (Map) it.next();
            String documentId = (String) map.get(DOCUMENT_ID);
            wordSourceIds.add(getWordSourceId(conn, documentId));
        }

        return wordSourceIds;
    }

    /**
     * <p>
     * Updates the use count of table DOCUMENT according to the given document index id and use count.
     * </p>
     *
     * <p>
     * Note, this operation relates to table DOCUMENT.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param documentIndexId the unique string id of WordSourceId instance
     * @param newUseCount the new use count to set for the document
     * @throws SQLException when exception occurs during the database operation
     */
    static void updateDocumentUsageCount(Connection conn,
        String documentIndexId, int newUseCount) throws SQLException {
        List args = new LinkedList();
        args.add(new Integer(newUseCount));
        args.add(documentIndexId);

        // update the USE_COUNT column in table DOCUMENT
        DBUtil.executeUpdate(conn, UPDATE_DOCUMENT, args);
    }

    /**
     * <p>
     * Adds or removes the relationship between document and collection document.
     * Which action to execute depends on the database and the given CollectionIndex instance.
     * </p>
     *
     * <p>
     * Note, this operation relates to table DOCUMENT_COLLECTION_DOCUMENT_XREF.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param collectionIndex the unique string id of collection document
     * @throws SQLException when exception occurs during the database operation
     * @throws IndexPersistenceException when fails to generate unique string id for WordSourceId instance
     */
    static void updateCollectionIndex(Connection conn,
        CollectionIndex collectionIndex)
        throws SQLException, IndexPersistenceException {
        String collectionId = collectionIndex.getId();

        // select all the relationships
        List result = DBUtil.executeQuery(conn,
                SELECT_DOCUMENT_COLLECTION_DOCUMENT_XREF,
                Collections.singletonList(collectionId));

        // this set stores all the document ids that have relationship with the given document collection
        // in database
        Set persistedDocumentIds = new HashSet();

        for (Iterator it = result.iterator(); it.hasNext();) {
            Map map = (Map) it.next();
            persistedDocumentIds.add(map.get(DOCUMENT_ID));
        }

        // this set stores all the document ids that have relationship with the given document collection
        // in memory
        Set documentIds = new HashSet();

        for (Iterator it = collectionIndex.getAllDocumentIds().iterator();
                it.hasNext();) {
            WordSourceId wordSourceId = (WordSourceId) it.next();
            documentIds.add(DocumentIndexPersistenceUtil.getDigestBasedId(
                    wordSourceId));
        }

        // insert the relationships according to documentIds
        for (Iterator it = documentIds.iterator(); it.hasNext();) {
            String documentIndexId = (String) it.next();

            if (!persistedDocumentIds.contains(documentIndexId)) {
                updateDocumentCollectionDocumentXref(conn, collectionId, documentIndexId, true);
            }
        }

        // delete the relationships according to documentIds
        for (Iterator it = persistedDocumentIds.iterator(); it.hasNext();) {
            String documentIndexId = (String) it.next();

            if (!documentIds.contains(documentIndexId)) {
                updateDocumentCollectionDocumentXref(conn, collectionId, documentIndexId, false);
            }
        }
    }

    /**
     * <p>
     * Adds or removes the relationship between document and collection document.
     * Which action to execute depends on the given flag.
     * </p>
     *
     * <p>
     * Note, this operation relates to table DOCUMENT_COLLECTION_DOCUMENT_XREF.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param collectionId the unique string id of collection document
     * @param documentIndexId the unique string id of collection document
     * @param isAdded a flag to identify whether it's an add action or a remove action
     * @throws SQLException when exception occurs during the database operation
     */
    static void updateDocumentCollectionDocumentXref(Connection conn,
        String collectionId, String documentIndexId, boolean isAdded)
        throws SQLException {
        List xrefArgs = new LinkedList();
        xrefArgs.add(collectionId);
        xrefArgs.add(documentIndexId);

        if (isAdded) {
            DBUtil.executeUpdate(conn,
                INSERT_DOCUMENT_COLLECTION_DOCUMENT_XREF, xrefArgs);
        } else {
            DBUtil.executeUpdate(conn,
                DELETE_DOCUMENT_COLLECTION_DOCUMENT_XREF, xrefArgs);
        }
    }

    /**
     * <p>
     * Creates a new DBConnectionFactory instance from values configed in the given namespace.
     * </p>
     *
     * @param namespace a namespace specifies where to get the values needed
     * @return the new DBConnectionFactory instance
     * @throws PersistenceConfigurationException when the creation of DBConnectionFactory instance fails.
     */
    static DBConnectionFactory getDBConnectionFactory(String namespace)
        throws PersistenceConfigurationException {
        //get the configuration value from connectionFactoryClass property
        //it is optional.
        String connectionFactoryClass = DocumentIndexPersistenceUtil.getString(namespace,
                CONNECTION_FACTORY_CLASS, false);

        //get the configuration value from connectionFactoryNamespace property
        //it is required.
        String connectionFactoryNamespace = DocumentIndexPersistenceUtil.getString(namespace,
                CONNECTION_FACTORY_NAMESPACE, true);

        if (connectionFactoryClass != null) {
            //use reflection to create DBConnectionFactory instance
            Class[] parametersType = new Class[] {String.class};
            Object[] parametersValue = new Object[] {connectionFactoryNamespace};

            return (DBConnectionFactory) DocumentIndexPersistenceUtil.createObject(DBConnectionFactory.class,
                connectionFactoryClass, parametersType, parametersValue);
        } else {
            // the class name is not specified, so use the default factory implementation class
            // reflection is not needed
            try {
                return new DBConnectionFactoryImpl(connectionFactoryNamespace);
            } catch (UnknownConnectionException e) {
                throw new PersistenceConfigurationException("UnknownConnectionException occurs when creating "
                        + "DBConnectionFactoryImpl instance.", e);
            } catch (ConfigurationException e) {
                throw new PersistenceConfigurationException("ConfigurationException occurs when creating "
                        + "DBConnectionFactoryImpl instance.", e);
            }
        }
    }

    /**
     * <p>
     * This function is used to execute a query sql expression in database persistence.
     * </p>
     *
     * <p>
     * Each record returned by database are wrapped in a HashMap instance, the key is column name string,
     * the value is the record value(in Object form).
     * All the records are wrapped as a list, in which the element is HashMap instance.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param sql the query sql expression
     * @param param the parameters for executing query in database
     * @return a List contains all the records queried from database.
     * @throws SQLException when exception occurs during the database operation
     */
    static List executeQuery(Connection conn, String sql, List param)
        throws SQLException {
        List retList = new ArrayList();
        ResultSet rs = null;
        ResultSetMetaData rsmeta;

        PreparedStatement ps = conn.prepareStatement(sql);

        // set up all the necessary parameters for executing.
        for (int i = 0; i < param.size(); i++) {
            setElement(i + 1, param.get(i), ps);
        }

        rs = ps.executeQuery();
        rsmeta = rs.getMetaData();

        // stores all the column names in returned ResultSet.
        List columnList = new ArrayList();

        int columnNum = rsmeta.getColumnCount();
        for (int i = 1; i <= columnNum; i++) {
            columnList.add(rsmeta.getColumnName(i));
        }

        while (rs.next()) {
            Map hash = new HashMap();

            for (int i = 0; i < columnNum; i++) {
                String columName = ((String) columnList.get(i)).toUpperCase();

                // the current selected record will be wrapped in a HashMap instance
                // the key is the column name, the value is corresponding object
                hash.put(columName, rs.getObject(columName));
            }

            retList.add(hash);
        }

        // release database resource
        closeResultSet(rs);
        closeStatement(ps);

        return retList;
    }

    /**
     * <p>
     * This function is used to execute an update sql expression in database persistence.
     * </p>
     *
     * @param conn the connection instance for database operation.
     * @param sql the update sql expression
     * @param param the parameters for executing update in database
     * @return the number of rows affects by the current operation
     * @throws SQLException when exception occurs during the database operation
     */
    static int executeUpdate(Connection conn, String sql, List param)
        throws SQLException {
        PreparedStatement ps = conn.prepareStatement(sql);

        // set up all the necessary parameters for executing.
        for (int i = 0; i < param.size(); i++) {
            setElement(i + 1, param.get(i), ps);
        }

        int result = ps.executeUpdate();

        // release database resource
        closeStatement(ps);

        return result;
    }

    /**
     * <p>
     * Sets the actual value to replace the corresponding question mark.
     * </p>
     *
     * @param order the sequence number of question mark in sql expression
     * @param parameter the actual value to replace the corresponding question mark.
     * @param ps PreparedStatement instance to execute the sql expression
     * @throws SQLException when exception occurs during the database operation
     */
    static void setElement(int order, Object parameter, PreparedStatement ps)
        throws SQLException {
        // replace the question mark in sql with real value
        if (parameter instanceof String) {
            ps.setString(order, (String) parameter);
        } else if (parameter instanceof Integer) {
            ps.setInt(order, ((Integer) parameter).intValue());
        } else if (parameter instanceof Long) {
            ps.setLong(order, ((Long) parameter).longValue());
        } else if (parameter instanceof byte[]) {
            // it's used for CLOB and BLOB column
            byte[] data = (byte[]) parameter;
            InputStream is = new ByteArrayInputStream(data);
            ps.setBinaryStream(order, is, (int) (data.length));
        } else {
            throw new IllegalArgumentException(
                "The given parameter type is not supported.");
        }
    }

    /**
     * <p>
     * Closes the given Connection instance.
     * </p>
     *
     * @param con the given Connection instance to close.
     */
    static void closeConnection(Connection con) {
        try {
            if ((con != null) && !con.isClosed()) {
                con.close();
            }
        } catch (SQLException e) {
            // Ignore
        }
    }

    /**
     * <p>
     * Closes the given PreparedStatement instance.
     * </p>
     *
     * @param state the given PreparedStatement instance to close.
     */
    public static void closeStatement(PreparedStatement state) {
        try {
            if (state != null) {
                state.close();
            }
        } catch (SQLException e) {
            // Ignore
        }
    }

    /**
     * <p>
     * Closes the given ResultSet instance.
     * </p>
     *
     * @param rs the given ResultSet instance to close.
     */
    public static void closeResultSet(ResultSet rs) {
        try {
            if (rs != null) {
                rs.close();
            }
        } catch (SQLException e) {
            // Ignore
        }
    }
}
