/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.DBConnectionFactory;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.DocumentIndexPersistenceUtil;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.sql.Connection;
import java.sql.SQLException;

import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * <p>
 * This class extends AbstractDBIndexPersistence class, which implements IndexPersistence interface.
 * It is called as simple access which means a fast, CLOB based, normalized data based schema.
 * </p>
 *
 * <p>
 * In this solution, it provides for a user who is concerned about speed of the execution. By storing all the word and
 * word position data in a single CLOB structure we are achieving a much speed increase over a comparable solution that
 * uses thousands of records to store words and related positions.
 * </p>
 *
 * <p>
 * Note, this component is acted as a plug-in component for Document Indexer component.
 * This class implements the persistence layer, the interface contract is defined by Document Indexer component.
 * </p>
 *
 * <p>
 * Important, In Document Indexer component, WordSourceId instance is used as an identifier for a DocumentIndex
 * instance, but for storing purpose, an complex object is not suit for primary key and so the WordSourceId will be
 * converted to a unique string id using MD5 algorithm before persisted.
 * </p>
 *
 * <p>
 * Thread Safety : the implementation is not thread-safe, since all the methods interact with the persistent storage.
 * However, all the remove, add and update methods, transaction is used, and if any error occurs during removal or add,
 * transaction will be rollbacked. It's clear that this class makes no guarantee of externally thread safe.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 *
 */
public class FastAccessDBIndexPersistence extends AbstractDBIndexPersistence {
    /**
     * <p>
     * Constructs a new FastAccessDBIndexPersistence instance.
     * This constructor will populate the connectionFactory and connectionName information from configuration.
     * </p>
     *
     * @param namespace The namespace for connection-related properties.
     * @throws IllegalArgumentException if namespace param is null or empty.
     * @throws PersistenceConfigurationException if there are configuration issues encountered
     */
    public FastAccessDBIndexPersistence(String namespace)
        throws PersistenceConfigurationException {
        super(namespace);
    }

    /**
     * <p>
     * Constructs a new FastAccessDBIndexPersistence instance.
     * This constructor will populate the connectionFactory and connectionName information from input parameters.
     * </p>
     *
     * @param connectionFactory DBConnectionFactory instance used to generate connections.
     * @param connectionName The name of the connection used for persistence operations.
     * @throws IllegalArgumentException if connectionFactory is null.
     */
    public FastAccessDBIndexPersistence(DBConnectionFactory connectionFactory,
        String connectionName) {
        super(connectionFactory, connectionName);
    }

    /**
     * <p>
     * Adds the given DocumentIndex to the persistence.
     * </p>
     *
     * <p>
     * Note, the SourceIdentity will be serialized before persisting.
     * In this solution, the words data will be stored as a CLOB column for speed purpose.
     * It's an error situation when the given documentIndex has already been persisted.
     * </p>
     *
     * @param documentIndex WordSource representing the document to be indexed, should not be null
     * @throws IllegalArgumentException when documentIndex is null
     * @throws IndexPersistenceException when fails to add DocumentIndex to the persistence or the
     * DocumentIndex has been persisted already.
     */
    public void addDocumentIndex(DocumentIndex documentIndex)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(documentIndex, "documentIndex");

        Connection conn = null;

        try {
            conn = getConnection();

            WordSourceId wordSourceId = documentIndex.getWordSourceId();
            String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

            // query the DOCUMENT table for existence
            Document document = DBUtil.queryDocument(conn, documentIndexId);

            if (document == null) {
                beginTransaction(conn);

                try {
                    DBUtil.insertDocument(conn, documentIndex, documentIndexId);
                    DBUtil.fastInsertWords(conn, documentIndex, documentIndexId);
                    commitTransaction(conn);
                } catch (SQLException e) {
                    rollbackTransaction(conn);
                    throw new IndexPersistenceException("SQLException occurs when adding document index.", e);
                } catch (IndexPersistenceException e) {
                    rollbackTransaction(conn);
                    throw new IndexPersistenceException("SQLException occurs when adding document index.", e);
                }
            } else {
                throw new IndexPersistenceException(
                    "The document index exists in persistence.");
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when adding document index.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Retrieves a DocumentIndex with the specified WordSourceId from the persistence.
     * If DocumentIndex with the given WordSourceId is not found in the persistence, null is expected.
     * </p>
     *
     * <p>
     * This method will not take the situation that the given WordSourceId is not found in persistence as a mistake.
     * In that case, null will be returned.
     * </p>
     *
     * <p>
     * In this solution, the words is stored as a CLOB column so the retrieve speed is much faster compared to
     * the simple access solution.
     * </p>
     *
     * @return retrieved DocumentIndex, or null if document index with given WordSourceId is not found in the
     * persistence
     * @param wordSourceId WordSourceId of document to retrieve, should not be null
     * @throws IllegalArgumentException when wordSourceId is null
     * @throws IndexPersistenceException when fails to retrieve document index with the specified WordSourceId
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        Connection conn = null;

        try {
            conn = getConnection();

            String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

            // query the DOCUMENT table for existence
            Document document = DBUtil.queryDocument(conn, documentIndexId);

            if (document == null) {
                return null;
            } else {
                // this is a fast solution
                Map words = DBUtil.fastQueryWords(conn, documentIndexId, wordSourceId);

                return new DocumentIndex(wordSourceId, words);
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when getting document index.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Removes a document index with given WordSourceId from the persistence.
     * This method will only remove the DocumentIndex with zero use count, if it's not satisfied,
     * then exception will be issued.
     * </p>
     *
     * <p>
     * When the document index is not found, or it fails to remove it exception is thrown.
     * </p>
     *
     * <p>
     * Document index use count is an integer value assigned to every document, initially it's zero.
     * If use count is not zero, it mean that document index is added
     * to one or more document collection indices, and so it can't be removed.
     * all the modification to use count is done by increaseUseCount() and decreaseUseCount() methods.
     * </p>
     *
     * @param wordSourceId WordSourceId of  document index to remove, should not be null
     * @throws IllegalArgumentException when WordSourceId is null
     * @throws IndexPersistenceException when document index is not found in the persistence, or error happens when
     * trying to remove it
     */
    public void removeDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        Connection conn = null;

        try {
            conn = getConnection();

            String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

            //query the DOCUMENT table for existence
            Document document = DBUtil.queryDocument(conn, documentIndexId);

            if (document == null) {
                throw new IndexPersistenceException(
                    "The given document index doesn't exist in persistence when removing.");
            } else {
                int useCount = document.getUseCount();

                // only document index with zero use count can be removed
                if (useCount != 0) {
                    throw new IndexPersistenceException(
                        "The use count of the removing document index is not zero.");
                }

                beginTransaction(conn);

                try {
                    DBUtil.removeDocumentIndex(conn, documentIndexId, false);
                    commitTransaction(conn);
                } catch (SQLException e) {
                    rollbackTransaction(conn);
                    throw new IndexPersistenceException("SQLException occurs when removing document index.",
                        e);
                }
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when removing document index.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Stores document collection index in the persistence.
     * </p>
     *
     * <p>
     * Note, all the relationships between documents and this document collection will be stored as well.
     * It's an error situation when the given collectionIndex has already been persisted.
     * </p>
     *
     * @param collectionIndex CollectionIndex to store, should not be null
     * @throws IllegalArgumentException if collectionIndex is null
     * @throws IndexPersistenceException if fails to create and store document collection index, or the document
     * collection index has been persisted already.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(collectionIndex,
            "collectionIndex");

        Connection conn = null;

        try {
            conn = getConnection();

            //query the DOCUMENT_COLLECTION_DOCUMENT table for existence
            if (DBUtil.isCollectionIndexPersisted(conn, collectionIndex.getId())) {
                throw new IndexPersistenceException(
                    "The collection index exists in persistence.");
            }

            beginTransaction(conn);

            try {
                DBUtil.insertCollectionIndex(conn, collectionIndex);
                commitTransaction(conn);
            } catch (SQLException e) {
                rollbackTransaction(conn);
                throw new IndexPersistenceException("SQLException occurs when adding collection index.",
                    e);
            } catch (IndexPersistenceException e) {
                rollbackTransaction(conn);
                throw e;
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when adding collection index.", e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Retrieves a document collection index with specified identifier.
     * Null will returned to identify collection with specified identifier is not found in persistence.
     * In this method, all the associated document index will be retrieved from persistence as well.
     * </p>
     *
     * <p>
     * This method will not take the situation that the given collection id is not found in persistence as a mistake.
     * In that case, null will be returned.
     * Otherwise, throw exception when collection index cannot be retrieved.
     * </p>
     *
     * <p>
     * Note, when creating the CollectionIndex object using the three-param constructor, the owner IndexManager
     * param will be set to null.
     * </p>
     *
     * @return CollectionIndex with specified identifier
     * @param collectionId identifier of document collection index to retrieve, should not be null or empty string
     * @throws IllegalArgumentException when collectionId is null or empty string
     * @throws IndexPersistenceException when collection index can not be retrieved
     */
    public CollectionIndex getCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkString(collectionId, "collectionId");

        Connection conn = null;

        try {
            conn = getConnection();

            // query the DOCUMENT_COLLECTION_DOCUMENT table for existence
            if (!DBUtil.isCollectionIndexPersisted(conn, collectionId)) {
                return null;
            } else {
                return DBUtil.getCollectionIndex(conn, collectionId, false);
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when getting collection index.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Removes document collection index with specified identifier from the persistence.
     * It removes the document collection record, all the relationships between documents and this document collection,
     * decreases the use count of corresponding documents as well.
     * </p>
     *
     * <p>
     * It will throw exception when there is no CollectionIndex with specified identifier found in persistence.
     * </p>
     *
     * @param collectionId Identifier of document  collection index to remove, should not be null or empty string
     * @throws IllegalArgumentException when collectionId is null or empty string
     * @throws IndexPersistenceException when fails to remove CollectionInde with given identifier or the document
     * collection record doesn't exist in database
     */
    public void removeCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkString(collectionId, "collectionId");

        Connection conn = null;

        try {
            conn = getConnection();

            //query the DOCUMENT_COLLECTION_DOCUMENT table for existence
            if (!DBUtil.isCollectionIndexPersisted(conn, collectionId)) {
                throw new IndexPersistenceException(
                    "The given collection index doesn't exist in persistence when removing.");
            } else {
                List documents = DBUtil.getDocuments(conn, collectionId);
                beginTransaction(conn);

                try {
                    DBUtil.removeCollectionIndex(conn, collectionId, documents);
                    commitTransaction(conn);
                } catch (SQLException e) {
                    rollbackTransaction(conn);
                    throw new IndexPersistenceException("SQLException occurs when removing collection index.",
                        e);
                }
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when removing collection index.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Updates specified CollectionIndex in the persistence.
     * The relationships between documents and the document collection will be updated.
     * Add or remove action to execute will depends on the difference between documents in persistence and
     * documents in the given instance, both of which associate with the collection index id.
     * </p>
     *
     * <p>
     * It will throw exception when there is no CollectionIndex with specified identifier found in persistence.
     * </p>
     *
     * <p>
     * Note, the use count will be updated in Document Indexer component, which calls decreaseUseCount() or
     * increaseUseCount() instead.
     * </p>
     *
     * @param collectionIndex CollectionIndex to update, should not be null
     * @throws IllegalArgumentException if collectionIndex parameter is null
     * @throws IndexPersistenceException if any error happens when updating the collection index in the persistence,
     * this includes the case, when the specified collectionIndex is not found in the persistence
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(collectionIndex,
            "collectionIndex");

        Connection conn = null;

        try {
            conn = getConnection();

            //query the DOCUMENT_COLLECTION_DOCUMENT table for existence
            if (!DBUtil.isCollectionIndexPersisted(conn, collectionIndex.getId())) {
                throw new IndexPersistenceException(
                    "The given collection index doesn't exist in persistence when removing.");
            } else {
                beginTransaction(conn);

                try {
                    DBUtil.updateCollectionIndex(conn, collectionIndex);
                    commitTransaction(conn);
                } catch (SQLException e) {
                    rollbackTransaction(conn);
                    throw new IndexPersistenceException("SQLExcetion occurs when updating collection index.",
                        e);
                }
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLExcetion occurs when updating collection index.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Increases the use count value for document index with specified WordSourceId.
     * Use count is the integer  value stored in the persistence along with every document index.
     * This value is equal to the number of collections, that contain a particular document.
     * </p>
     *
     * <p>
     * It will throw exception when the WordSourceId cannot be found in persistence.
     * </p>
     *
     * @param wordSourceId WordSourceId of document index use count of which to update, should not be null
     * @throws IllegalArgumentException if wordSourceId is null
     * @throws IndexPersistenceException when fails to increase document index use count in the persistence
     * or it is not found in persistence
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);
        Connection conn = null;

        try {
            conn = getConnection();

            //query the DOCUMENT table for existence
            Document document = DBUtil.queryDocument(conn, documentIndexId);

            if (document == null) {
                throw new IndexPersistenceException(
                    "The given word source id doesn't exist in persistence when increasing use count.");
            } else {
                beginTransaction(conn);

                int userCount = document.getUseCount() + 1;

                try {
                    DBUtil.updateDocumentUsageCount(conn, documentIndexId,
                        userCount);
                    commitTransaction(conn);
                } catch (SQLException e) {
                    rollbackTransaction(conn);
                    throw new IndexPersistenceException("SQLException occurs when increasing document use count.",
                        e);
                }
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLExcetion occurs when increasing collection index.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Decreases the use count value for document index with specified WordSourceId.
     * Use count is the integer  value stored in the persistence along with every document index.
     * This value is equal to the number of collections, that contain a particular document.
     * </p>
     *
     * <p>
     * It will throw exception when the WordSourceId cannot be found in persistence.
     * </p>
     *
     * @param wordSourceId WordSourceId of document index use count of which to update, should not be null
     * @throws IllegalArgumentException if wordSourceId is null
     * @throws IndexPersistenceException when fails to increase document index use count in the persistence
     * or it is not found in persistence
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);
        Connection conn = null;

        try {
            conn = getConnection();

            //query the DOCUMENT table for existence
            Document document = DBUtil.queryDocument(conn, documentIndexId);

            if (document == null) {
                throw new IndexPersistenceException(
                    "The given word source id doesn't exist in persistence when increasing use count.");
            } else {
                beginTransaction(conn);

                int userCount = document.getUseCount() - 1;

                try {
                    DBUtil.updateDocumentUsageCount(conn, documentIndexId, userCount);
                    commitTransaction(conn);
                } catch (SQLException e) {
                    rollbackTransaction(conn);
                    throw new IndexPersistenceException("SQLException occurs when increasing document use count.",
                        e);
                }
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLExcetion occurs when increasing collection index.", e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Return all the WordSourceId of documents that have document index persisted.
     * </p>
     *
     * <p>
     * Note, element in returned set is type of WordSourceId.
     * </p>
     *
     * @return set of WordSourceId of documents that have document index persisted.
     * @throws IndexPersistenceException when fails to retrieved all the WordSourceId set
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        Connection conn = null;

        try {
            conn = getConnection();

            return DBUtil.getIndexedDocuments(conn);
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when getting indexed documents.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }

    /**
     * <p>
     * Return the count of how many different groups this document is a member of.
     * </p>
     *
     * @return the count of how many different groups this document is a member of.
     * @param wordSourceId WordSourceId of document index use count of which to update, should not be null
     * @throws IllegalArgumentException if wordSourceId is null
     * @throws IndexPersistenceException when fails to get document index use count in the persistence
     */
    public int getDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        Connection conn = null;

        try {
            conn = getConnection();

            String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);
            Document document = DBUtil.queryDocument(conn, documentIndexId);

            //query the DOCUMENT table for existence
            if (document == null) {
                throw new IndexPersistenceException(
                    "The given word source id doesn't exist in persistence when getting use count.");
            } else {
                return document.getUseCount();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("SQLException occurs when getting use count.",
                e);
        } finally {
            DBUtil.closeConnection(conn);
        }
    }
}
