/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * <p>
 * This class extends DefaultHandler and is used to parse xml content.
 * It is used by SAX parser and is only used parse document usage xml.
 * </p>
 *
 * <p>
 * Note, SAX is used for loading xml file at the sake of speed and memory compared to DOM based parser.
 * </p>
 *
 * <p>
 * The following is a sample of document collection xml.
 * &lt;?xml version=&quot;1.0&quot;?&gt;
 * &lt;documents&gt;
 *     &lt;document-usage document-id=&quot;16572&quot; usage-count=&quot;2&quot;&gt;
 *         &lt;collection-inclusion&gt;
 *             &lt;collection collection-id=&quot;37684&quot;/&gt;
 *             &lt;collection collection-id=&quot;11111&quot;/&gt;
 *         &lt;/collection-inclusion&gt;
 *     &lt;/document-usage&gt;
 *     &lt;document-usage document-id=&quot;46537&quot; usage-count=&quot;0&quot;/&gt;
 * &lt;/documents&gt;
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 *
 */
final class DocumentUsageHandler extends DefaultHandler {
    /**
     * <p>
     * A map instance used to store the mapping from document id to its corresponding use count.
     * The key is type of String while the value is type of Integer.
     * </p>
     */
    private Map documentUsageMap;

    /**
     * <p>
     * A map instance used to store the mapping from document id to its corresponding collection id list.
     * The key is type of String while the value is type of List, in which element is type of String.
     * </p>
     */
    private Map documentUsageCountMap;

    /**
     * <p>
     * Represents all the collection ids for the current document index id.
     * </p>
     */
    private List collectionList;

    /**
     * <p>
     * Represents the current document index id processing by SAX Parsing.
     * </p>
     */
    private String currentDocumentId;

    /**
     * <p>
     * Creates a new DocumentUsageHandler instance.
     * </p>
     *
     * <p>
     * Note, it's package private.
     * </p>
     */
    DocumentUsageHandler() {
        documentUsageMap = new HashMap();
        documentUsageCountMap = new HashMap();
    }

    /**
     * <p>
     * When the name of the current element is document-usage, collection, the corresponding document-id, usage-count
     * and collection-id attributes are parsed.
     * </p>
     *
     * <p>
     * Note, when the SAX parser encounters a star-tag, this method is called.
     * </p>
     *
     * @param namespaceURI namespace uri
     * @param localName local name
     * @param qName qualified name
     * @param attr represents the attributes for the current element encountered
     *
     * @throws SAXException Any SAX exception, possibly wrapping another exception.
     */
    public void startElement(String namespaceURI, String localName,
        String qName, Attributes attr) throws SAXException {
        String currentElementName = qName.intern();

        if (currentElementName == Constants.DOCUMENT_USAGE) {
            // when document-usage element is encountered, then save documen-id attribute
            collectionList = new ArrayList();
            currentDocumentId = attr.getValue(Constants.DOCUMENT_ID);

            String count = attr.getValue(Constants.USAGE_COUNT);
            documentUsageCountMap.put(currentDocumentId, Integer.valueOf(count));
        } else if (currentElementName == Constants.COLLECTION) {
            // when collection element is encountered, then save collection-id attribute
            collectionList.add(attr.getValue(Constants.COLLECTION_ID));
        }
    }

    /**
     * <p>
     * If document-usage element ends, then currentDocumentId and the corresponding collectionList is stored to a map.
     * </p>
     *
     * <p>
     * Note, when the SAX parser encounters a end-tag, this method is called.
     * </p>
     *
     * @param namespaceURI namespace uri
     * @param localName local name
     * @param qName qualified name
     *
     * @throws SAXException Any SAX exception, possibly wrapping another exception.
     */
    public void endElement(String namespaceURI, String localName, String qName)
        throws SAXException {
        if (qName.intern() == Constants.DOCUMENT_USAGE) {
            // when the end tag of document-usage element is encountered, then the usage of
            // a document index should be complete.
            documentUsageMap.put(currentDocumentId, collectionList);
        }
    }

    /**
     * <p>
     * Returns the usage count that relates to a given document index id.
     * </p>
     *
     * @param documentId the given document index id to get use count
     * @return the usage count that relates to a given document index id.
     */
    Integer getDocumentUsageCount(String documentId) {
        return (Integer) documentUsageCountMap.get(documentId);
    }

    /**
     * <p>
     * Returns the collection id list that relates to a given document index id.
     * </p>
     *
     * @param documentId the given document index id to get associated collection id list.
     * @return the collection id list that relates to a given document index id.
     */
    List getDocumentCollectionList(String documentId) {
        return (List) documentUsageMap.get(documentId);
    }
}
