/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.DocumentIndexPersistenceUtil;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;

import java.text.CollationKey;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;


/**
 * <p>
 * This class extends AbstractDBIndexPersistence class, which implements IndexPersistence interface.
 * </p>
 *
 * <p>
 * This implementation uses xml files as storage medium a dual strategy for parsing.
 * It utilizes SAX parsing for reading data in and DOM parsing for data updates.
 * This implementation allows for creation of arbitrarily large document index files by sequencing
 * (splitting) large files into chunks of file based data. Since it is required that binary data is
 * stored in the xml files for document index a serialized Base64 encoded data will be both encoded
 * and decoded by this implementation.
 * </p>
 *
 * <p>
 * Note, this component is acted as a plug-in component for Document Indexer component.
 * This class implements the persistence layer, the interface contract is defined by Document Indexer component.
 * </p>
 *
 * <p>
 * Important, In Document Indexer component, WordSourceId instance is used as an identifier for a DocumentIndex
 * instance, but for storing purpose, an complex object is not suit for primary key and so the WordSourceId will be
 * converted to a unique string id using MD5 algorithm before persisted.
 * </p>
 *
 * <p>
 * For this implementation, documet_usage.xml is considered as part of configuration, it must exist in the persistence
 * directory.
 * Here an initial document_usage.xml is shown as below:
 * &lt;?xml version=\&quot;1.0\&quot;?&gt;
 * &lt;documents&gt;
 * &lt;/documents&gt;
 * </p>
 *
 * <p>
 * Thread Safety : the implementation is not thread-safe, since all the methods interact with the xml file storage,
 * it makes no guarantee thread safe.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public class XmlIndexPersistence implements IndexPersistence {
    /**
     * <p>
     * The buffer size constant for constructing a string buffer instance.
     * </p>
     */
    private static final int BUFFER_SIZE = 1024;

    /**
     * <p>
     * The constant to be used when converting from MB to KB, or KB to Byte.
     * </p>
     */
    private static final int UNIT = 1024;

    /**
     * <p>
     * Represents the FileSizeLimit property name.
     * </p>
     */
    private static final String FILESIZE_LIMIT = "FileSizeLimit";

    /**
     * <p>
     * Represents the FileSizeThreshold property name.
     * </p>
     */
    private static final String FILESIZE_THRESHOLD = "FileSizeThreshold";

    /**
     * <p>
     * Represents the xmlPersistencePath property name.
     * </p>
     */
    private static final String XML_PERSISTENCE_PATH = "xmlPersistencePath";

    /**
     * <p>
     * This represent the file name used to represent a document usage as defined by the persistence
     * contract. There is only one such file per xml persistence name space.
     * </p>
     */
    private static final String DOC_USAGE_FILE_NAME = "document_usage.xml";

    /**
     * <p>
     * This represent the file name prefix for all files that represent a Collection Index xml file
     * storage (i.e. a document collection file)
     * Each group xml file name will have the following format: doc_group_groupId_.xml
     * </p>
     *
     * <p>
     * For example here is a simple document group xml file for a group with id of 175658:
     * doc_collection_175658.xml
     * </p>
     */
    private static final String DOC_COLLECTION_FILE_NAME_PREFIX = "doc_collection_";

    /**
     * <p>
     * This represent the file name prefix for all files that represent a Document Index xml file storage
     * (i.e. a document index file)
     * Each document index file name will have the following format: doc_index_docId_sequenceid.xml
     * </p>
     *
     * <p>
     * The sequence id is used to link files with data that is larger than say a limit of 2G (this can be
     * configured) this means that if we have indexed data that is larger than say 2G we split the file into
     * 2 files with the first having the sequence id of 1 and the second one being obviously 2. This gives the
     * flexibility for persistence of very large content.
     * Note, if the sequence ids are not consecutive then it's consider as a mistake.
     * For example, sequence ids such as "1, 2, 4, 5" obviously mean that the document index files are destroyed.
     * </p>
     *
     * <p>
     * Here is an example of a document index xml file for a document with id of 16527:
     *       doc_index_16527_1.xml
     * </p>
     */
    private static final String DOC_INDEX_FILE_NAME_PREFIX = "doc_index_";

    /**
     * <p>
     * Represents the length of beginning section of document index file.
     * It is used to calculate the current file length when generating document index file.
     * </p>
     */
    private static final long BEGINING_LENGTH = Constants.XML_HEADER.getBytes().length
        + Constants.DOC_START.getBytes().length + Constants.DOCUMENTINDEX_START.getBytes().length;

    /**
     * <p>
     * Represents the length of ending section of document index file.
     * It is used to calculate the current file length when generating document index file.
     * </p>
     */
    private static final long ENDING_LENGTH = Constants.DOCUMENTINDEX_END.getBytes().length
        + Constants.DOC_END.getBytes().length;

    /**
     * <p>
     * This is a file size limit for saving a chunk of indexed data. It is measured in Mbytes.
     * Consider data so large that it cannot fit into a single file due to size limitations for the specific OS.
     * This size limit is used to decide when to start chaining multiple files to store the data.
     * It is initialized from configuration through the constructor and once initialized cannot be changed.
     * Must be a positive integer (i.e. &gt; 0)
     * </p>
     */
    private final int fileSizeLimit;

    /**
     * <p>
     * This is a safety net size in Kbytes that is applied before the  fileSizeLimit is reached.
     * This is basically something of a round off error precaution.
     * It is initialized from configuration through the constructor and once initialized cannot be changed.
     * Must be a non-negative integer (i.e. &gt;= 0)
     * </p>
     */
    private final int fileSizeThreshold;

    /**
     * <p>
     * This is the file path to a directory where all the xml files that make up this persistence session
     * are stored (it is almost like a data base namespace). This must be a valid fully qualified path or relative
     * path.
     * It is initialized from configuration through the constructor and once initialized cannot be changed.
     * Cannot be null or an empty string.
     * </p>
     */
    private final String persistenceDirPath;

    /**
     * <p>
     * Constructs a new XmlIndexPersistence instance.
     * This constructor will populate the fileSizeLimit, fileSizeThreshold, and persistenceDirPath
     * information from configuration.
     * </p>
     *
     * @param namespace configuration namespace
     * @throws PersistenceConfigurationException if there are configuration issues encountered, which includes
     * that the persistenceDirPath doesn't represents a existant, writable directory.
     * @throws IllegalArgumentException if namespace param is null or empty.
     */
    public XmlIndexPersistence(String namespace)
        throws PersistenceConfigurationException {
        DocumentIndexPersistenceUtil.checkString(namespace, "namespace");

        String fileSizeLimitValue = DocumentIndexPersistenceUtil.getString(namespace,
                FILESIZE_LIMIT, true);

        fileSizeLimit = DocumentIndexPersistenceUtil.checkInt(fileSizeLimitValue,
                "fileSizeLimit", "GT", 0);

        String fileSizeThresholdValue = DocumentIndexPersistenceUtil.getString(namespace,
                FILESIZE_THRESHOLD, true);

        fileSizeThreshold = DocumentIndexPersistenceUtil.checkInt(fileSizeThresholdValue,
                "fileSizeThreshold", "GE", 0);

        persistenceDirPath = DocumentIndexPersistenceUtil.getString(namespace,
                XML_PERSISTENCE_PATH, true);

        if (!checkPersistenceDirPath()) {
            throw new PersistenceConfigurationException(
                "The persistenceDirPath doesn't represents a existant, writable directory.");
        }
    }

    /**
     * <p>
     * Constructs a new XmlIndexPersistence instance.
     * This constructor will populate the fileSizeLimit, fileSizeThreshold, and persistenceDirPath information directly.
     * </p>
     *
     * @param fileSizeLimit This is a file size limit for saving a chunk of indexed data
     * @param fileSizeThreshold This is a safety net size in Kbytes that is applied before the fileSizeLimit is reached
     * @param persistenceDirPath This is the file path to a directory where all the xml files that make up this
     * persistence session are stored
     * @throws IllegalArgumentException if persistenceDirPath param is null or empty, or persistenceDirPath doesn't
     * represents a existant, writable directory. or fileSizeLimit is less than or equal to zero, or fileSizeThrehold
     * is less than zero.
     */
    public XmlIndexPersistence(int fileSizeLimit, int fileSizeThreshold,
        String persistenceDirPath) {
        DocumentIndexPersistenceUtil.checkInt(fileSizeLimit, "fileSizeLimit",
            "GT", 0);
        DocumentIndexPersistenceUtil.checkInt(fileSizeThreshold,
            "fileSizeThreshold", "GE", 0);
        DocumentIndexPersistenceUtil.checkString(persistenceDirPath,
            "persistenceDirPath");

        this.fileSizeLimit = fileSizeLimit;
        this.fileSizeThreshold = fileSizeThreshold;
        this.persistenceDirPath = persistenceDirPath;

        if (!checkPersistenceDirPath()) {
            throw new IllegalArgumentException(
                "The persistenceDirPath doesn't represents a existant, writable directory.");
        }
    }

    /**
     * <p>
     * Adds the given DocumentIndex to the persistence.
     * </p>
     *
     * <p>
     * Note, the SourceIdentity will be serialized before persisting.
     * </p>
     *
     * @param documentIndex WordSource representing the document to be indexed, should not be null
     * @throws IllegalArgumentException when documentIndex is null
     * @throws IndexPersistenceException when fails to add DocumentIndex to the persistence
     */
    public void addDocumentIndex(DocumentIndex documentIndex)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(documentIndex, "documentIndex");

        // the sequence id will be start with one, and increased one by one.
        int sequenceId = 1;

        // the file size limit, in byte unit.
        long limitLength = (fileSizeLimit * UNIT * UNIT) - (fileSizeThreshold * UNIT);

        String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(documentIndex.getWordSourceId());

        String filePrefix = DOC_INDEX_FILE_NAME_PREFIX + documentIndexId + "_";

        if (listIndexFiles(filePrefix).size() != 0) {
            throw new IndexPersistenceException("The given document index id exist in persistence when adding.");
        }

        StringBuffer docBuffer = new StringBuffer(BUFFER_SIZE);

        StringBuffer wordIndexBuffer = new StringBuffer();
        StringBuffer wordSourceDataBuffer = generateWordSourceData(documentIndex);

        generateHeaderBuffer(documentIndex, documentIndexId, docBuffer, wordSourceDataBuffer);

        long currentFileLength = BEGINING_LENGTH
            + wordSourceDataBuffer.toString().getBytes().length + ENDING_LENGTH;

        long wordIndexLength = 0;

        Map wordsMap = documentIndex.getWords();

        for (Iterator it = wordsMap.entrySet().iterator(); it.hasNext();) {
            Entry entry = (Entry) it.next();
            CollationKey key = (CollationKey) entry.getKey();
            List wordPosList = (List) entry.getValue();

            StringBuffer currentWordIndexBuffer = generateCurrentWordIndex(key, wordPosList);
            wordIndexBuffer.append(currentWordIndexBuffer);

            wordIndexLength += currentWordIndexBuffer.toString().getBytes().length;

            // if the size current file length is greater than limit, the write it into file system
            if ((currentFileLength + wordIndexLength) >= limitLength) {

                String fileName = filePrefix + sequenceId + ".xml";
                writeIndexFile(new File(persistenceDirPath, fileName),
                    docBuffer, wordIndexBuffer);
                wordIndexBuffer.delete(0, wordIndexBuffer.length());
                wordIndexLength = 0;
                sequenceId++;
            }
        }

        if (wordIndexBuffer.length() != 0) {
            String fileName = filePrefix + sequenceId + ".xml";
            writeIndexFile(new File(persistenceDirPath, fileName), docBuffer, wordIndexBuffer);
        }

        // create a document-usage element in document-usage.xml
        createDocumentUsage(documentIndexId);
    }

    /**
     * <p>
     * Retrieves a DocumentIndex with the specified WordSourceId from the persistence.
     * If DocumentIndex with the given WordSourceId is not found in the persistence, null is expected.
     * </p>
     *
     * <p>
     * This method will not take the situation that the given WordSourceId is not found in persistence as a mistake.
     * In that case, null will be returned.
     * </p>
     *
     * @return retrieved DocumentIndex, or null if document index with given WordSourceId is not found in the
     * persistence
     * @param wordSourceId WordSourceId of document to retrieve, should not be null
     * @throws IllegalArgumentException when wordSourceId is null
     * @throws IndexPersistenceException when fails to retrieve document index with the specified WordSourceId
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

        return retrieveDocumentIndex(wordSourceId, documentIndexId);
    }

    /**
     * <p>
     * Removes a document index with given WordSourceId from the persistence.
     * This method will only remove the DocumentIndex with zero use count, if it's not satisfied,
     * then exception will be issued.
     * </p>
     *
     * <p>
     * When the document index is not found, or it fails to remove it exception is thrown.
     * </p>
     *
     * <p>
     * Document index use count is an integer value assigned to every document, initially it's zero.
     * If use count is not zero, it mean that document index is added
     * to one or more document collection indices, and so it can't be removed.
     * all the modification to use count is done by increaseUseCount() and decreaseUseCount() methods.
     * </p>
     *
     * @param wordSourceId WordSourceId of document index to remove, should not be null
     * @throws IllegalArgumentException when WordSourceId is null
     * @throws IndexPersistenceException when document index is not found in the persistence, or error happens
     * when trying to remove it
     */
    public void removeDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

        String filePrefix = DOC_INDEX_FILE_NAME_PREFIX + documentIndexId + "_";
        Map indexFilesMap = listIndexFiles(filePrefix);

        if (indexFilesMap.size() == 0) {
            throw new IndexPersistenceException(
                "The given document index id doesn't exist in persistence when removing.");
        }

        File usageFile = getUsageFile();

        // remove the document-usage element in document-usage.xml
        // note, only the document index with zero use count will be removed, otherwise it's a mistake
        Document usageDoc = parseDOMDocument(usageFile);
        NodeList nodes = usageDoc.getElementsByTagName(Constants.DOCUMENT_USAGE);
        for (int i = 0; i < nodes.getLength(); i++) {
            Element usageElement = (Element) nodes.item(i);

            if (documentIndexId.equals(usageElement.getAttribute(
                            Constants.DOCUMENT_ID))) {
                int userCount = Integer.parseInt(usageElement.getAttribute(
                            Constants.USAGE_COUNT));

                if (userCount != 0) {
                    throw new IndexPersistenceException(
                        "The use count of the removing document index is not zero.");
                } else {
                    usageElement.getParentNode().removeChild(usageElement);
                    writeDOMDocument(usageFile, usageDoc);
                    break;
                }
            }
        }

        // remove all the index files
        for (int i = 0; i < indexFilesMap.size(); i++) {
            File file = (File) indexFilesMap.get(filePrefix + (i + 1) + ".xml");

            if (file == null) {
                throw new IndexPersistenceException(
                    "The document index is not consistency, index file with number "
                    + (i + 1) + " was lost.");
            } else {
                file.delete();
            }
        }
    }

    /**
     * <p>
     * Stores document collection index in the persistence.
     * </p>
     *
     * <p>
     * Note, all the relationships between documents and this document collection will be stored as well.
     * It's an error situation when the given collectionIndex has already been persisted.
     * </p>
     *
     * @param collectionIndex CollectionIndex to store, should not be null
     * @throws IllegalArgumentException if collectionIndex is null
     * @throws IndexPersistenceException if fails to create and store document collection index, or the
     * document collection index has been persisted already.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(collectionIndex, "collectionIndex");

        File file = getCollectionIndexFile(collectionIndex.getId());

        if (file.exists()) {
            throw new IndexPersistenceException("The collection index exists in persistence.");
        }

        // write a new document collection xml file
        Document doc = createBlankDOMDocument();
        Element documentCollection = doc.createElement(Constants.DOCUMENT_COLLECTION);
        documentCollection.setAttribute(Constants.COLLECTION_ID,
            collectionIndex.getId());

        Element documents = doc.createElement(Constants.DOCUMENTS);
        Set wordSourceIds = collectionIndex.getAllDocumentIds();

        for (Iterator it = wordSourceIds.iterator(); it.hasNext();) {
            WordSourceId wordSourceId = (WordSourceId) it.next();
            String documentId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

            //update the use count
            File usageFile = getUsageFile();
            Document usageDoc = parseDOMDocument(usageFile);
            updateUsageCountDOM(usageDoc, documentId, true);
            updateUsageDOM(usageDoc, collectionIndex.getId(), documentId, true);
            writeDOMDocument(usageFile, usageDoc);

            Element documentIdElement = doc.createElement(Constants.DOCUMENT_ID);
            Text text = doc.createTextNode(documentId);
            documentIdElement.appendChild(text);
            documents.appendChild(documentIdElement);
        }

        documentCollection.appendChild(documents);
        doc.appendChild(documentCollection);
        writeDOMDocument(file, doc);
    }

    /**
     * <p>
     * Retrieves a document collection index with specified identifier.
     * Null will returned to identify collection with specified identifier is not found in persistence.
     * In this method, all the associated document index will be retrieved from persistence as well.
     * </p>
     *
     * <p>
     * This method will not take the situation that the given collection id is not found in persistence as a mistake.
     * In that case, null will be returned.
     * Otherwise, throw exception when collection index cannot be retrieved.
     * </p>
     *
     * <p>
     * Note, when creating the CollectionIndex object using the three-param constructor, the owner IndexManager
     * param will be set to null.
     * </p>
     *
     * @return CollectionIndex with specified identifier
     * @param collectionId identifier of document collection index to retrieve, should not be null or empty string
     * @throws IllegalArgumentException when collectionId is null or empty string
     * @throws IndexPersistenceException when collection index can not be retrieved
     */
    public CollectionIndex getCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkString(collectionId, "collectionId");

        File file = getCollectionIndexFile(collectionId);

        // null is expected when the collection id hasn't been persisted
        if (!file.exists()) {
            return null;
        }

        CollectionIndexHandler handler = new CollectionIndexHandler();
        parseSAXDocument(file, handler);

        Map wordsMap = new HashMap();
        Set wordSourceIds = new HashSet();

        for (Iterator it = handler.getDocumentIdList().iterator();
                it.hasNext();) {
            String documentId = (String) it.next();

            // the complete document index will be retrieved.
            DocumentIndex documentIndex = retrieveDocumentIndex(null, documentId);

            WordSourceId wordSourceId = documentIndex.getWordSourceId();
            wordSourceIds.add(wordSourceId);

            // converts the words in document index to be suitable for document collection
            for (Iterator iter = documentIndex.getWords().keySet().iterator();
                    iter.hasNext();) {
                CollationKey key = (CollationKey) iter.next();
                Set documentIds = (Set) wordsMap.get(key);

                if (documentIds == null) {
                    // The word does not exist, create a new list
                    documentIds = new HashSet();
                    wordsMap.put(key, documentIds);
                }

                // Mark the word's existence in the document
                documentIds.add(wordSourceId);
            }
        }

        return new CollectionIndex(null, wordsMap, wordSourceIds, collectionId);
    }

    /**
     * <p>
     * Removes document collection index with specified identifier from the persistence.
     * It removes the document collection record, all the relationships between documents and this document collection,
     * decreases the use count of corresponding documents as well.
     * </p>
     *
     * <p>
     * It will throw exception when there is no CollectionIndex with specified identifier found in persistence.
     * </p>
     *
     * @param collectionId Identifier of document  collection index to remove, should not be null or empty string
     * @throws IllegalArgumentException when collectionId is null or empty string
     * @throws IndexPersistenceException when fails to remove CollectionInde with given identifier or the document
     * collection record doesn't exist in persistence
     */
    public void removeCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkString(collectionId, "collectionId");

        File file = getCollectionIndexFile(collectionId);

        if (!file.exists()) {
            throw new IndexPersistenceException("The collectionIndex id hasn't been persisted.");
        }

        CollectionIndexHandler handler = new CollectionIndexHandler();
        parseSAXDocument(file, handler);

        File usageFile = new File(persistenceDirPath, DOC_USAGE_FILE_NAME);
        Document doc = parseDOMDocument(usageFile);

        //update usage
        for (Iterator it = handler.getDocumentIdList().iterator();
                it.hasNext();) {
            String documentIndexId = (String) it.next();
            updateUsageDOM(doc, collectionId, documentIndexId, false);
        }

        file.delete();
    }

    /**
     * <p>
     * Updates specified CollectionIndex in the persistence.
     * The relationships between documents and the document collection will be updated.
     * Add or remove action to execute will depends on the difference between documents in persistence and
     * documents in the given instance, both of which associate with the collection index id.
     * </p>
     *
     * <p>
     * It will throw exception when there is no CollectionIndex with specified identifier found in persistence.
     * </p>
     *
     * <p>
     * Note, the use count will be updated in Document Indexer component, which calls decreaseUseCount() or
     * increaseUseCount() instead.
     * </p>
     *
     * @param collectionIndex CollectionIndex to update, should not be null
     * @throws IllegalArgumentException if collectionIndex parameter is null
     * @throws IndexPersistenceException if any error happens when updating the collection index in the persistence,
     * this includes the case, when the specified collectionIndex is not found in the persistence
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(collectionIndex, "collectionIndex");

        String collectionId = collectionIndex.getId();

        File file = getCollectionIndexFile(collectionId);

        if (!file.exists()) {
            throw new IndexPersistenceException("The collectionIndex instance hasn't been persisted.");
        }

        Document doc = parseDOMDocument(file);

        File usageFile = getUsageFile();
        Document usageDoc = parseDOMDocument(usageFile);

        // this set stores all the document ids that have relationship with the given document collection
        // in xml persistence root
        Set persistedDocumentIds = getDocumentIndices(collectionId);

        // this set stores all the document ids that have relationship with the given document collection
        // in memory
        Set documentIds = new HashSet();
        for (Iterator it = collectionIndex.getAllDocumentIds().iterator();
                it.hasNext();) {
            WordSourceId wordSourceId = (WordSourceId) it.next();
            documentIds.add(DocumentIndexPersistenceUtil.getDigestBasedId(
                    wordSourceId));
        }

        // insert the relationships according to documentIds
        for (Iterator it = documentIds.iterator(); it.hasNext();) {
            String documentIndexId = (String) it.next();

            if (!persistedDocumentIds.contains(documentIndexId)) {
                updateCollectionDOM(doc, collectionId, documentIndexId, true);
                // this insert the collection element indocument-usage.xml
                // the use count will be updated in increaseDocumentUseCount()
                updateUsageDOM(usageDoc, collectionId, documentIndexId, true);
            }
        }

        // delete the relationships according to documentIds
        for (Iterator it = persistedDocumentIds.iterator(); it.hasNext();) {
            String documentIndexId = (String) it.next();

            if (!documentIds.contains(documentIndexId)) {
                updateCollectionDOM(doc, collectionId, documentIndexId, false);
                // this remove the collection element in document-usage.xml
                // the use count will be updated in decreaseDocumentUseCount()
                updateUsageDOM(usageDoc, collectionId, documentIndexId, false);
            }
        }

        writeDOMDocument(file, doc);
        writeDOMDocument(usageFile, usageDoc);
    }

    /**
     * <p>
     * Increases the use count value for document index with specified WordSourceId.
     * Use count is the integer  value stored in the persistence along with every document index.
     * This value is equal to the number of collections, that contain a particular document.
     * </p>
     *
     * <p>
     * It will throw exception when the WordSourceId cannot be found in persistence.
     * </p>
     *
     * @param wordSourceId WordSourceId of document index use count of which to update, should not be null
     * @throws IllegalArgumentException if wordSourceId is null
     * @throws IndexPersistenceException when fails to increase document index use count in the persistence
     * or it is not found in persistence
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

        File usageFile = getUsageFile();
        Document doc = parseDOMDocument(usageFile);
        updateUsageCountDOM(doc, documentIndexId, true);
        writeDOMDocument(usageFile, doc);
    }

    /**
     * <p>
     * Decreases the use count value for document index with specified WordSourceId.
     * Use count is the integer  value stored in the persistence along with every document index.
     * This value is equal to the number of collections, that contain a particular document.
     * </p>
     *
     * <p>
     * It will throw exception when the WordSourceId cannot be found in persistence.
     * </p>
     *
     * @param wordSourceId WordSourceId of document index use count of which to update, should not be null
     * @throws IllegalArgumentException if wordSourceId is null
     * @throws IndexPersistenceException when fails to increase document index use count in the persistence
     * or it is not found in persistence
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        String documentIndexId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

        File usageFile = getUsageFile();
        Document doc = parseDOMDocument(usageFile);
        updateUsageCountDOM(doc, documentIndexId, false);
        writeDOMDocument(usageFile, doc);
    }

    /**
     * <p>
     * Return all the WordSourceId of documents that have document index persisted.
     * </p>
     *
     * <p>
     * Note, element in returned set is type of WordSourceId.
     * </p>
     *
     * @return set of WordSourceId of documents that have document index persisted.
     * @throws IndexPersistenceException when fails to retrieved all the WordSourceId set
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        Set wordSourceIds = new HashSet();
        FileFilter filter = new XMLPersistenceFileFilter(DOC_INDEX_FILE_NAME_PREFIX, "_1.xml");
        File[] indexFiles = new File(persistenceDirPath).listFiles(filter);

        for (int i = 0; i < indexFiles.length; i++) {
            DocumentIndexHandler handler = new DocumentIndexHandler(null);
            parseSAXDocument(indexFiles[i], handler);
            wordSourceIds.add(handler.getWordSourceId());
        }

        return wordSourceIds;
    }

    /**
     * <p>
     * Return the count of how many different groups this document is a member of.
     * </p>
     *
     * @return the count of how many different groups this document is a member of.
     * @param wordSourceId WordSourceId of document index use count of which to update, should not be null
     * @throws IllegalArgumentException if wordSourceId is null
     * @throws IndexPersistenceException when fails to get document index use count in the persistence
     */
    public int getDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        DocumentIndexPersistenceUtil.checkNull(wordSourceId, "wordSourceId");

        String documentId = DocumentIndexPersistenceUtil.getDigestBasedId(wordSourceId);

        String filePrefix = DOC_INDEX_FILE_NAME_PREFIX + documentId + "_";
        Map indexFilesMap = listIndexFiles(filePrefix);
        if (indexFilesMap.size() == 0) {
            throw new IndexPersistenceException(
                    "The given word source id doesn't exist in persistence when getting use count.");
        }

        DocumentUsageHandler handler = new DocumentUsageHandler();
        parseSAXDocument(getUsageFile(), handler);

        return handler.getDocumentUsageCount(documentId).intValue();
    }

    /**
     * <p>
     * Searches the persistence directory to return a map, in which the key is file name string,
     * the value is File instance.
     * All the files returned will be part of a DocumentIndex instance.
     * </p>
     *
     * <p>
     * The selected file must start with filePrefix and ends with ".xml", which represents part of
     * document index file.
     * </p>
     *
     * @param filePrefix the file prefix of a document index file
     * @return a index files map, in which the key is file name string, the value is File instance. All the files
     * returned will be part of a DocumentIndex instance.
     */
    private Map listIndexFiles(String filePrefix) {
        // Searches the persistence directory for all the files that start with the given filePrefix and ends
        // with ".xml"
        FileFilter filter = new XMLPersistenceFileFilter(filePrefix, ".xml");
        File[] indexFiles = new File(persistenceDirPath).listFiles(filter);
        Map indexFilesMap = new HashMap();

        // Saves the result Files in a Map, the key is file name while the value is File instance
        for (int i = 0; i < indexFiles.length; i++) {
            indexFilesMap.put(indexFiles[i].getName(), indexFiles[i]);
        }

        return indexFilesMap;
    }

    /**
     * <p>
     * Returns a DocumentIndex instance for a specified document index id querying from xml file persistence.
     * </p>
     *
     * <p>
     * The pass-in wordSourceId can be null or not, when it's not null, it means we can ignore the wordSourceId
     * parts when parsing xml file.
     * </p>
     *
     * @param wordSourceId the WordSourceId instance, may be null or not null
     * @param documentIndexId the unique string id of a DocumentIndex instance, not null
     * @return a DocumentIndex instance querying from xml file persistence
     * @throws IndexPersistenceException when sequence ids are not consecutive, or failed to deserialize source
     * identity object
     */
    private DocumentIndex retrieveDocumentIndex(WordSourceId wordSourceId,
        String documentIndexId) throws IndexPersistenceException {
        String filePrefix = DOC_INDEX_FILE_NAME_PREFIX + documentIndexId + "_";
        Map indexFilesMap = listIndexFiles(filePrefix);

        // if the DocumentIndex is not persisted, null is expected
        if (indexFilesMap.size() == 0) {
            return null;
        }

        Map words = new HashMap();

        // reads all the index files
        for (int i = 0; i < indexFilesMap.size(); i++) {
            File file = (File) indexFilesMap.get(filePrefix + (i + 1) + ".xml");

            if (file == null) {
                throw new IndexPersistenceException(
                    "The document index is not consistency, index file with number "
                    + (i + 1) + " was lost.");
            } else {

                DocumentIndexHandler handler = new DocumentIndexHandler(wordSourceId);
                parseSAXDocument(file, handler);
                words.putAll(handler.getWords());

                if (wordSourceId == null) {
                    wordSourceId = handler.getWordSourceId();
                }
            }
        }

        return new DocumentIndex(wordSourceId, words);
    }

    /**
     * <p>
     * Returns the File instance represents the usage xml file.
     * </p>
     *
     * <p>
     * Note, if the usage file can't be found in the persistence directory, then it will throw
     * IndexPersistenceException.
     * </p>
     *
     * @return the File instance represents the usage xml file.
     * @throws IndexPersistenceException when usage file doesn't exist in persistence directory
     */
    private File getUsageFile() throws IndexPersistenceException {
        File usageFile = new File(persistenceDirPath, DOC_USAGE_FILE_NAME);

        if (!usageFile.exists()) {
            throw new IndexPersistenceException(
                "The document_usage.xml doesn't exist in the persistence path.");
        }

        return usageFile;
    }

    /**
     * <p>
     * Returns a set of documents that associate with the document collection specified by the given collection id.
     * In the returned set, the element is type of String.
     * </p>
     *
     * @param collectionId the unique string id for document collection
     * @return a set of documents that associate with the document collection specified by the given collection id.
     * @throws IndexPersistenceException when exception occurs when operating the xml file.
     */
    private Set getDocumentIndices(String collectionId)
        throws IndexPersistenceException {
        File file = getCollectionIndexFile(collectionId);

        // parse the collection index file to get all the relationships
        CollectionIndexHandler handler = new CollectionIndexHandler();
        parseSAXDocument(file, handler);

        Set set = new HashSet();

        for (Iterator it = handler.getDocumentIdList().iterator(); it.hasNext();) {
            set.add(it.next());
        }

        return set;
    }

    /**
     * <p>
     * Generates the word-source-data element from the given DocumentIndex instance and then returns
     * in form of StringBuffer.
     * </p>
     *
     * @param documentIndex DocumentIndex instance to save to xml persistence
     * @return the StringBuffer instance that contains the word-source-data element which is part of document index file
     * @throws IndexPersistenceException when failed to encode source identity object
     */
    private StringBuffer generateWordSourceData(DocumentIndex documentIndex)
        throws IndexPersistenceException {
        WordSourceId wordSourceId = documentIndex.getWordSourceId();

        StringBuffer wordSourceDataBuffer = new StringBuffer(BUFFER_SIZE);
        wordSourceDataBuffer.append(Constants.WORDSOURCEDATA_START);

        // first serialize the source identity, then encode the byte array using BASE64 algorithm
        String identity = DocumentIndexPersistenceUtil.base64Encode(
                DocumentIndexPersistenceUtil.serializeObject(wordSourceId.getSourceIdentity()));

        wordSourceDataBuffer.append(Constants.SOURCEIDENTITY_ELEMENT.replaceFirst(
                Constants.PLACE_HOLDER, identity));
        wordSourceDataBuffer.append(Constants.CLASSNAME_ELEMENT.replaceFirst(
                Constants.PLACE_HOLDER, wordSourceId.getSourceClassName()));

        // generate the locale element
        Locale locale = documentIndex.getLocale();
        String country = locale.getCountry();
        String language = locale.getLanguage();
        String variant = locale.getVariant();
        String localeAttrs = "language=\"" + language + "\" country=\""
            + country + "\" variant=\"" + variant + "\"";
        wordSourceDataBuffer.append(Constants.LOCALE_ELEMENT.replaceFirst(
                Constants.PLACE_HOLDER, localeAttrs));

        wordSourceDataBuffer.append(Constants.DELIMITERS_START);

        // all the delimiters will be saved in their hex string representation
        String[] delimiters = documentIndex.getWordSourceId().getDelimiters();
        String[] hexCodedDelimiters = new String[delimiters.length];

        for (int i = 0; i < delimiters.length; i++) {
            hexCodedDelimiters[i] = DocumentIndexPersistenceUtil.hexCode(delimiters[i]);
        }

        for (int i = 0; i < delimiters.length; i++) {
            wordSourceDataBuffer.append(Constants.ENTRY_ELEMENT.replaceFirst(
                    Constants.PLACE_HOLDER, hexCodedDelimiters[i]));
        }

        wordSourceDataBuffer.append(Constants.DELIMITERS_END);
        wordSourceDataBuffer.append(Constants.WORDSOURCEDATA_END);

        return wordSourceDataBuffer;
    }

    /**
     * <p>
     * Generates the header part of a new document index file.
     * </p>
     *
     * @param documentIndex documentIndex instance
     * @param documentIndexId unique string id of DocumentIndex instance
     * @param docBuffer whole buffer for the new document index file
     * @param wordSourceDataBuffer the buffer for word-source-data element
     */
    private void generateHeaderBuffer(DocumentIndex documentIndex, String documentIndexId, StringBuffer docBuffer,
            StringBuffer wordSourceDataBuffer) {
        docBuffer.append(Constants.XML_HEADER);
        docBuffer.append(Constants.DOC_START.replaceFirst(
                Constants.PLACE_HOLDER, documentIndex.getLocale().toString()));
        docBuffer.append(Constants.DOCUMENTINDEX_START.replaceFirst(
                Constants.PLACE_HOLDER, documentIndexId));

        // add the word-source-data element
        docBuffer.append(wordSourceDataBuffer);
    }

    /**
     * <p>
     * Creates a new document-usage element in document-usage xml file.
     * </p>
     *
     * @param documentIndexId unique string id of DocumentIndex instance
     * @throws IndexPersistenceException when failed to create a new document-usage element in document-usage xml file.
     */
    private void createDocumentUsage(String documentIndexId) throws IndexPersistenceException {
        File usageFile = getUsageFile();
        Document usageDoc = parseDOMDocument(usageFile);
        NodeList nodes = usageDoc.getElementsByTagName(Constants.DOCUMENTS);
        Element element = (Element) nodes.item(0);

        // create documents element in case that it's not found
        if (element == null) {
            element = usageDoc.createElement(Constants.DOCUMENTS);
            usageDoc.appendChild(element);
        }

        // create a new document-usage element
        Element usageElement = usageDoc.createElement(Constants.DOCUMENT_USAGE);
        usageElement.setAttribute(Constants.DOCUMENT_ID, documentIndexId);
        usageElement.setAttribute(Constants.USAGE_COUNT, "0");
        element.appendChild(usageElement);
        writeDOMDocument(usageFile, usageDoc);
    }

    /**
     * <p>
     * Generates the word-index element from the given collation key and word position list and then returns
     * in form of StringBuffer.
     *
     * @param key the representation of a word
     * @param wordPosList the position list
     * @return the current word index buffer
     */
    private StringBuffer generateCurrentWordIndex(CollationKey key, List wordPosList) {
        StringBuffer currentWordIndexBuffer = new StringBuffer(BUFFER_SIZE);
        currentWordIndexBuffer.append(Constants.WORDINDEX_START);

        // generate word element
        currentWordIndexBuffer.append(Constants.WORD_ELEMENT.replaceFirst(
                Constants.PLACE_HOLDER, key.getSourceString()));

        for (Iterator pIt = wordPosList.iterator(); pIt.hasNext();) {
            Integer pos = (Integer) pIt.next();

            // generate all the pos elements for the current word element
            currentWordIndexBuffer.append(Constants.POS_ELEMENT.replaceFirst(
                    Constants.PLACE_HOLDER, String.valueOf(pos)));
        }

        currentWordIndexBuffer.append(Constants.WORDINDEX_END);
        return currentWordIndexBuffer;
    }

    /**
     * <p>
     * Returns the File instance represents the document collection file.
     * </p>
     *
     * @param collectionId the unique string id for document collection
     * @return the File instance represents the document collection file.
     */
    private File getCollectionIndexFile(String collectionId) {
        File file = new File(persistenceDirPath,
                DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + ".xml");

        return file;
    }

    /**
     * <p>
     * Returns true if the persitenceDirPath represents a existant, writable directory, otherwise false.
     * </p>
     *
     * @return true if the persitenceDirPath represents a existant, writable directory, otherwise false.
     */
    private boolean checkPersistenceDirPath() {
        File dir = new File(this.persistenceDirPath);

        if (!dir.exists()) {
            return false;
        }

        if (!dir.isDirectory()) {
            return false;
        }

        if (!dir.canWrite()) {
            return false;
        }

        return true;
    }

    /**
     * <p>
     * Writes a document index file to the specified File instance.
     * Note, only the header part and the current word index buffer is specified because the the ending
     * of the document index file is fixed, it is saved as a constant defined in Constants class.
     * </p>
     *
     * @param file the place where the document index file writes to
     * @param header the header buffer of the document index file
     * @param currentWordIndexBuffer the current word index buffer
     * @throws IndexPersistenceException when IOException occurs in writing index file
     */
    private void writeIndexFile(File file, StringBuffer header,
        StringBuffer currentWordIndexBuffer) throws IndexPersistenceException {
        Writer output = null;

        try {
            FileOutputStream fos = new FileOutputStream(file);
            // UTF-8 encoding is used for consistency
            OutputStreamWriter out = new OutputStreamWriter(fos, "UTF-8");
            output = new BufferedWriter(out);

            // writes a index file
            output.write(header.toString());
            output.write(currentWordIndexBuffer.toString());
            output.write(Constants.DOCUMENTINDEX_END);
            output.write(Constants.DOC_END);
        } catch (IOException e) {
            throw new IndexPersistenceException("IOException happens when writing xml file.",
                e);
        } finally {
            //flush and close both "output" and its underlying writer
            if (output != null) {
                try {
                    output.close();
                } catch (IOException e) {
                    //ignore
                }
            }
        }
    }

    /**
     * <p>
     * Parsing the given xml file with the specified DefaultHandler instance using SAX Parser.
     * </p>
     *
     * @param xmlFile the given xml file to parse
     * @param handler DefaultHandler instance to handle the xml content
     * @throws IndexPersistenceException when fails to parsing the given xml file
     */
    private void parseSAXDocument(File xmlFile, DefaultHandler handler)
        throws IndexPersistenceException {
        // use the SAXParser bundled with JDK
        SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setNamespaceAware(false);
        factory.setValidating(false);

        try {
            SAXParser parser = factory.newSAXParser();
            parser.parse(xmlFile, handler);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("ParserConfigurationException occurs when parsing xml file using SAX.",
                e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("SAXException occurs when parsing xml file using SAX.",
                e);
        } catch (IOException e) {
            throw new IndexPersistenceException("IOException occurs when parsing xml file using SAX.",
                e);
        }
    }

    /**
     * <p>
     * Updates usage DOM Document to insert or delete a collection element identified by the given collection
     * id and collection id.
     * The given isAdded flag is to identity whether it's an add or removal action.
     * when it's true, add the special element otherwise remove the special element.
     * </p>
     *
     * @param doc DOM Document represents usage file
     * @param collectionId unique string id of document collection
     * @param documentId unique string id of document
     * @param isAdded flag to identify whether to add or remove
     * @throws IndexPersistenceException when inconsistency found in collection file
     */
    private void updateUsageDOM(Document doc, String collectionId,
        String documentId, boolean isAdded) throws IndexPersistenceException {
        NodeList nodeList = doc.getElementsByTagName(Constants.DOCUMENT_USAGE);

        // search the DOM to find the document-usage element with specified document id.
        for (int i = 0; i < nodeList.getLength(); i++) {
            Element element = (Element) nodeList.item(i);

            // search for the collection element with specified collection-id attribute
            if (documentId.equals(element.getAttribute(Constants.DOCUMENT_ID))) {
                NodeList collections = element.getElementsByTagName(Constants.COLLECTION);
                Element collectionElement = null;

                for (int j = 0; j < collections.getLength(); j++) {
                    Element subElement = (Element) collections.item(j);

                    if (collectionId.equals(subElement.getAttribute(Constants.COLLECTION_ID))) {
                        collectionElement = subElement;

                        break;
                    }
                }

                // create a new collection element
                if ((collectionElement == null) && isAdded) {
                    collectionElement = doc.createElement(Constants.COLLECTION);
                    collectionElement.setAttribute(Constants.COLLECTION_ID,
                        collectionId);

                    Element collectionInclusion = (Element)
                        element.getElementsByTagName(Constants.COLLECTION_INCLUSION).item(0);

                    // create a new collection-inclusion element in case that it's not found
                    if (collectionInclusion == null) {
                        collectionInclusion = doc.createElement(Constants.COLLECTION_INCLUSION);
                        element.appendChild(collectionInclusion);
                    }

                    collectionInclusion.appendChild(collectionElement);
                // delete the collection element
                } else if ((collectionElement != null) && !isAdded) {
                    collectionElement.getParentNode().removeChild(collectionElement);
                } else if ((collectionElement != null) && isAdded) {
                    throw new IndexPersistenceException(
                        "Failed to add the association because the collection id "
                        + "is already associated with the given document index id.");
                } else {
                    throw new IndexPersistenceException(
                        "Failed to add the association because the collection id "
                        + "isn't associated with the given document index id.");
                }

                return;
            }
        }

        // the document index id can't be found in document-usage.xml file
        throw new IndexPersistenceException(
            "The document index doesn't exist in usage file");
    }

    /**
     * <p>
     * Updates usage DOM Document to add or subtract the use count belongs to the given document id.
     * The given isAdded flag is to identity whether to add or subtract use count by one.
     * </p>
     *
     * @param doc DOM Document represents usage file
     * @param documentId unique string id of document
     * @param isAdded flag to identify whether it's an add or subtract use count action.
     * @throws IndexPersistenceException when the given document id can not be found in usage file,
     * or the use count will be negative after subtraction
     */
    private void updateUsageCountDOM(Document doc, String documentId,
        boolean isAdded) throws IndexPersistenceException {
        NodeList nodeList = doc.getElementsByTagName(Constants.DOCUMENT_USAGE);

        // search the DOM to find the document-usage element with specified document id.
        for (int i = 0; i < nodeList.getLength(); i++) {
            Element element = (Element) nodeList.item(i);

            if (documentId.equals(element.getAttribute(Constants.DOCUMENT_ID))) {
                int count = Integer.parseInt(element.getAttribute(Constants.USAGE_COUNT));

                if (isAdded) {
                    element.setAttribute(Constants.USAGE_COUNT, String.valueOf(++count));
                } else {
                    if (count <= 0) {
                        throw new IndexPersistenceException(
                            "The use count will be negative after subtraction, which is not practical.");
                    }

                    element.setAttribute(Constants.USAGE_COUNT, String.valueOf(--count));
                }
                return;
            }
        }

        // the document index id can't be found in document-usage.xml file
        throw new IndexPersistenceException(
                "The document index doesn't exist in usage file when updating use count.");
    }

    /**
     * <p>
     * Updates collection DOM Document to insert or delete a document-id element identified by the given collection id.
     * The given isAdded flag is to identity whether to insert or remove.
     * when it's true, insert the special element otherwise remove the special element.
     * </p>
     *
     * @param collection DOM Document represents collection file
     * @param collectionId unique string id of document collection
     * @param documentId unique string id of document index
     * @param isAdded flag to state whether to add or remove
     * @throws IndexPersistenceException when inconsistency found in collection file
     */
    private void updateCollectionDOM(Document collection, String collectionId,
        String documentId, boolean isAdded) throws IndexPersistenceException {
        NodeList documentIds = collection.getElementsByTagName(Constants.DOCUMENT_ID);
        Element idElement = null;

        // search the DOM to find the document-id element with specified document id.
        for (int i = 0; i < documentIds.getLength(); i++) {
            Element element = (Element) documentIds.item(i);

            if (documentId.equals(getValue(element))) {
                idElement = element;
            }
        }

        // create a new document-id element
        if ((idElement == null) && isAdded) {
            Element documentsElement = (Element) collection.getElementsByTagName(Constants.DOCUMENTS)
                                                           .item(0);
            idElement = collection.createElement(Constants.DOCUMENT_ID);

            Text text = collection.createTextNode(documentId);
            idElement.appendChild(text);
            documentsElement.appendChild(idElement);
        // delete the document-id element
        } else if ((idElement != null) && !isAdded) {
            idElement.getParentNode().removeChild(idElement);
        } else if ((idElement != null) && isAdded) {
            throw new IndexPersistenceException(
                "Failed to add the association because the collection id "
                + "is already associated with the given document index id.");
        } else {
            throw new IndexPersistenceException(
                "Failed to add the association because the collection id "
                + "isn't associated with the given document index id.");
        }
    }

    /**
     * <p>
     * Parsing the given xml file and returns a DOM Document using DOM Parser.
     * </p>
     *
     * @param xmlFile the given xml file to parse
     * @return a DOM Document instance which represents the given xml file
     * @throws IndexPersistenceException when fails to parsing the given xml file
     */
    private Document parseDOMDocument(File xmlFile)
        throws IndexPersistenceException {
        // Parse the xml file using DOM API
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(false);
        factory.setValidating(false);

        try {
            DocumentBuilder builder = factory.newDocumentBuilder();

            // parse the file
            return builder.parse(xmlFile);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("ParserConfigurationException occurs when parsing xml file using DOM.",
                e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("SAXException occurs when parsing xml file using DOM.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("IOException occurs when parsing xml file using DOM.", e);
        }
    }

    /**
     * <p>
     * Creates a blank DOM Document instance and returns it.
     * </p>
     *
     * @return a new blank DOM Document instance
     * @throws IndexPersistenceException when fails to creating a new DOM Document instance.
     */
    private Document createBlankDOMDocument() throws IndexPersistenceException {
        // Create a blank DOM Document using DOM API
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(false);
        factory.setValidating(false);

        try {
            DocumentBuilder builder = factory.newDocumentBuilder();

            return builder.newDocument();
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("ParserConfigurationException occurs when creating document using DOM.",
                e);
        }
    }

    /**
     * <p>
     * Writes the DOM document to a file destination.
     * </p>
     *
     * @param file the storage destination
     * @param document DOM document to write its content
     * @throws IndexPersistenceException when IOException, TransformerException or
     * TransformerFactoryConfigurationError occurs during transformation and writing process
     */
    private void writeDOMDocument(File file, Document document)
        throws IndexPersistenceException {
        Writer writer = null;

        try {
            // Converts the Document instance and writes to file destination
            Transformer tansformer = TransformerFactory.newInstance().newTransformer();
            tansformer.setOutputProperty(OutputKeys.INDENT, "yes");

            Source source = new DOMSource(document);
            writer = new FileWriter(file);
            tansformer.transform(source, new StreamResult(writer));
        } catch (IOException ioe) {
            throw new IndexPersistenceException("IOException occurs when writing dom document to file system.",
                ioe);
        } catch (TransformerException e) {
            throw new IndexPersistenceException("TransformerException occurs when writing dom document to file system.",
                e);
        } catch (TransformerFactoryConfigurationError e) {
            throw new IndexPersistenceException("TransformerFactoryConfigurationError occurs when writing dom document "
                    + "to file system.", e);
        } finally {
            //flush and close writer
            if (writer != null) {
                try {
                    writer.close();
                } catch (IOException e) {
                    //ignore
                }
            }
        }
    }

    /**
     * <p>
     * Gets the value of specified node.
     * </p>
     *
     * @param node the node to get value
     *
     * @return the text value of specified element or null if not found
     */
    private String getValue(Node node) {
        if (node instanceof Element) {
            // The first child text_node or cdata section node be treat as value of element
            Element element = (Element) node;
            NodeList children = element.getChildNodes();

            for (int i = 0; i < children.getLength(); ++i) {
                Node child = children.item(i);

                if ((child.getNodeType() == Node.TEXT_NODE)
                        || (child.getNodeType() == Node.CDATA_SECTION_NODE)) {
                    return child.getNodeValue();
                }
            }
        } else {
            return node.getNodeValue();
        }

        return null;
    }

    /**
     * <p>
     * This class implements the FileFilter interface.
     * It is used to select a set of files whose name starts with a specified string
     * and ends with another specified string.
     * The pattern for start string and end string are assigned in constructor.
     * </p>
     *
     * @author TCSDEVELOPER
     * @version 1.0
     */
    private static class XMLPersistenceFileFilter implements FileFilter {
        /**
         * <p>
         * The start string to select the files that their name starts with.
         * </p>
         */
        private final String start;

        /**
         * <p>
         * The end string to select the files that their name ends with.
         * </p>
         */
        private final String end;

        /**
         * <p>
         * Constructs a XMLPersistenceFileFilter instance with start and end string specified.
         * </p>
         *
         * @param start the string to start with
         * @param end the string to end with
         */
        public XMLPersistenceFileFilter(String start, String end) {
            this.start = start;
            this.end = end;
        }

        /**
         * <p>
         * Implements the method defined by FileFilter interface.
         * In this method, if the given file that its name starts with the start string and ends with the
         * end string, then it's accepted.
         * </p>
         *
         * @param file a File instance to see whether it's accepted by this FileFilter implementation.
         * @return true if it's accepted otherwise false.
         */
        public boolean accept(File file) {
            String fileName = file.getName();

            return (fileName.startsWith(start) && fileName.endsWith(end));
        }
    }
}
